# Copyright (c) 2026 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
Stack
-----

The ``drgn.helpers.linux.stack`` module contains helpers for working with Linux
kernel stacks. These helpers can provide additional information specific to the
kernel.
"""
from dataclasses import dataclass
import enum
from itertools import chain
import operator
from typing import Iterator, List, Sequence, Tuple, Union

from drgn import (
    Architecture,
    IntegerLike,
    Object,
    ObjectNotFoundError,
    Program,
    StackFrame,
    StackTrace,
    TypeKind,
    sizeof,
)
from drgn.cli import display_str
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.pid import find_task
from drgn.helpers.linux.sched import task_cpu

__all__ = (
    "StackKind",
    "StackSegment",
    "LinuxKernelStack",
    "kernel_stack_trace",
)


class StackKind(enum.Enum):
    """A ``StackKind`` represents a kind of context associated with a stack."""

    USER = 1
    """Stack frames associated with userspace"""

    TASK = 2
    """Stack frames associated with the kernel's thread stack"""

    IRQ = 3
    """Stack frames associated with IRQ handlers"""

    SOFTIRQ = 4
    """Stack frames associated with softirqs"""

    NMI = 5
    """Stack frames associated with NMI handlers (arch-specific)"""

    EXCEPTION = 6
    """Stack frames associated with other exceptions (arch-specific)"""

    UNKNOWN = 7
    """Stack frames which could not be categorized"""


@dataclass(frozen=True)
class StackSegment:
    """A subset of a stack trace associated with a context."""

    kind: StackKind
    """The kind of stack associated with this segment"""

    frames: Sequence[StackFrame]
    """Stack frames that are part of the segment"""

    def __iter__(self) -> Iterator[StackFrame]:
        return iter(self.frames)

    def __len__(self) -> int:
        return len(self.frames)

    def __getitem__(self, key: int) -> StackFrame:
        return self.frames[key]


@display_str
@dataclass(frozen=True)
class LinuxKernelStack:
    """
    A stack trace broken into segments associated with execution contexts.

    Stack traces for kernel programs may span many execution contexts, each one
    interrupting the previous one. For instance, a user task may be interrupted
    by the kernel handling its system call, which could in turn be interrupted
    by a hardware interrupt. A ``LinuxKernelStack`` groups the frames of a stack
    trace into :class:`StackSegments<StackSegment>`, each of which is associated
    with a :class:`StackKind`. It may also be iterated and indexed like a normal
    :class:`~drgn.StackTrace`.
    """

    prog: Program
    """The program associated with the stack trace"""

    segments: Sequence[StackSegment]
    """Segments of the stack trace"""

    def __str__(self) -> str:
        lines = []
        i = -1
        for seg in self.segments:
            for i, frame in enumerate(seg.frames, i + 1):
                source_info = ""
                try:
                    source_info = " (%s:%d:%d)" % frame.source()
                except LookupError:
                    pass
                lines.append(f"#{i:<2d} {frame.name}{source_info}")
            lines.append(f"  --- end {seg.kind.name} stack ---")
        return "\n".join(lines)

    def __iter__(self) -> Iterator[StackFrame]:
        return chain.from_iterable(map(iter, self.segments))

    def __len__(self) -> int:
        return sum(len(seg.frames) for seg in self.segments)

    def __getitem__(self, index: int) -> StackFrame:
        if index < 0:
            index += len(self)
        if index < 0:
            raise IndexError("index out of range")
        start = 0
        for segment in self.segments:
            if start <= index < start + len(segment.frames):
                return segment.frames[index - start]
            start += len(segment.frames)
        raise IndexError("index out of range")


def _real_frames(trace: StackTrace) -> Iterator[List[StackFrame]]:
    # Yield all inline frames associated with their "physical" stack frame in a
    # single list. This simplifies iteration later.
    current = []
    for frame in trace:
        current.append(frame)
        if not frame.is_inline:
            yield current
            current = []
    if current:
        yield current


def _x86_64_exception_stacks(
    prog: Program, cpu: int
) -> List[Tuple[int, int, StackKind]]:
    try:
        # The cpu_entry_area was introduced in ece614dcfd964 ("x86/mm/fixmap:
        # Generalize the GDT fixmap mechanism, introduce struct
        # cpu_entry_area"), and later in the same series, the field
        # "exception_stacks" was introduced. This was all part of the KPTI
        # (Kernel Page Table Isolation) effort, which removes most of the
        # kernel's memory mappings from userspace page tables. This was merged
        # in v4.15 and backported to v4.14.9 stable.
        #
        # Prior to this, the per-CPU variable `exception_stacks` was an array of
        # stacks for different exception modes, and it was used directly for
        # exception stacks. But since KPTI results in this memory being
        # unmapped, the cpu_entry_area was created in "fixmap" memory which *is*
        # part of the minimal userspace page tables. The `exception_stacks`
        # variable still exists and contains the backing memory for these
        # mappings. But the stack pointers we observe use the addresses for the
        # cpu_entry_area, so the exception_stacks variable isn't helpful for
        # categorizing stacks once KPTI is present.
        #
        # So, when we see the cpu_entry_area type, we can assume that the KPTI
        # patches are present, and we should use it to find exception stacks.
        # Prior to that, we can just rely on exception_stacks.
        cea_tp = prog.type("struct cpu_entry_area")
    except LookupError:
        estacks = per_cpu(prog["exception_stacks"], cpu)
        base = estacks.address_of_().value_()
        return [(base, base + sizeof(estacks), StackKind.EXCEPTION)]

    _CEA_START = 0xFFFFFE0000000000 + prog["PAGE_SIZE"].value_()
    try:
        # Since v6.2 97e3d26b5e5f3 ("x86/mm: Randomize per-cpu entry area")
        # the CPU entry area has randomized offsets.
        cea_offset = int(per_cpu(prog["_cea_offset"], cpu))
    except ObjectNotFoundError:
        # Prior, it was a plain CPU index.
        cea_offset = cpu
    assert cea_tp.size is not None
    cea_start = _CEA_START + cea_offset * cea_tp.size

    stacks = []

    # Try to determine the NMI stack range and put it into the first slot, so we
    # can be more precise.
    try:
        # Since v5.1, in commit 019b17b3ffe48 ("x86/exceptions: Add structs for
        # exception stacks"), the exception stacks are stored in structs, and we
        # can easily identify the NMI stack (and even specific exception stacks,
        # if we cared to do so).
        cea = Object(prog, cea_tp, address=cea_start)
        estacks = cea.estacks
        nmi_stack = estacks.NMI_stack
        nmi_start = int(nmi_stack.address_of_())
        stacks.append((nmi_start, nmi_start + sizeof(nmi_stack), StackKind.NMI))

    except AttributeError:
        # Prior to that, there's a big array containing each stack in the
        # cpu_entry_area. Without hard-coding it, we can't really determine the
        # NMI stack, so don't bother. Just lump it in with the rest of the
        # exception stacks.
        pass

    # The CPU entry area for the CPU contains all other exception stacks.
    stacks.append((cea_start, cea_start + cea_tp.size, StackKind.EXCEPTION))
    return stacks


def _x86_64_stack_ranges(task: Object, cpu: int) -> List[Tuple[int, int, StackKind]]:
    prog = task.prog_

    stacks = _x86_64_exception_stacks(prog, cpu)

    try:
        # Since v6.15, the hardirq stack pointer is its own percpu global.
        # Commit c6a0918072eaa ("x86/irq: Move irq stacks to percpu hot
        # section") put it there.
        #
        # Strangely, prior to commit d7b6d709a76a4 ("x86/percpu: Move irq_stack
        # variables next to current_task") from v6.2, the hardirq_stack_ptr
        # variable was also a percpu global. So this suffices for that case as
        # well.
        hardirq_top = int(per_cpu(prog["hardirq_stack_ptr"], cpu))
    except ObjectNotFoundError:
        try:
            # As said above, between 6.2 and 6.12, it was found in pcpu_hot.
            hardirq_top = int(per_cpu(prog["pcpu_hot"], cpu).hardirq_stack_ptr)
        except ObjectNotFoundError:
            # And prior to v5.2 (commit 758a2e3122284 ("x86/irq/64: Rename
            # irq_stack_ptr to hardirq_stack_ptr")) it was just irq_stack_ptr:
            hardirq_top = int(per_cpu(prog["irq_stack_ptr"], cpu))

    # x86_64 defines IRQ_STACK_SIZE the same way as THREAD_SIZE
    irq_stack_size = prog["THREAD_SIZE"].value_()
    stacks.append((hardirq_top - irq_stack_size, hardirq_top, StackKind.IRQ))

    # On x86_64, userspace memory addresses always have the MSB cleared. The
    # actual userspace address range is smaller. However, don't include the zero
    # page in this, as it's not a valid stack pointer.
    stacks.append((int(prog["PAGE_SIZE"]), (1 << 63), StackKind.USER))

    return stacks


def _aarch64_stack_ranges(task: Object, cpu: int) -> List[Tuple[int, int, StackKind]]:
    prog = task.prog_
    irq_stack_ptr = per_cpu(prog["irq_stack_ptr"], cpu).value_()
    # aarch64 defines IRQ_STACK_SIZE as THREAD_SIZE
    thread_size = prog["THREAD_SIZE"].value_()
    return [
        (irq_stack_ptr, irq_stack_ptr + thread_size, StackKind.IRQ),
    ]


def _aarch64_classify(_: Object, __: int, frames: List[StackFrame]) -> StackKind:
    # aarch64 supports memory tagging, where the upper byte of a user memory
    # address can be set to a tag. The canonical address is formed by
    # sign-extending bit 55 into bits 56-63. In our case, we just want to test
    # it to determine whether the address is userspace or kernel.
    try:
        if not frames[0].sp & (1 << 55):
            return StackKind.USER
    except LookupError:
        pass
    return StackKind.UNKNOWN


def _arm_stack_ranges(task: Object, cpu: int) -> List[Tuple[int, int, StackKind]]:
    prog = task.prog_
    # ARM can have different user/kernel splits: 1/3, 2/2, and 3/1. While _stext
    # is not directly at that split, it's very close, and we can just mask off
    # the top two bits to get what the actual split is without any offset.
    kernel_offset = prog.symbol("_stext").address & 0xC0000000
    ranges = [
        (prog["PAGE_SIZE"].value_(), kernel_offset, StackKind.USER),
    ]
    try:
        irq_stack_ptr = per_cpu(prog["irq_stack_ptr"], cpu).value_()
        # arm allocates IRQ stacks at THREAD_SIZE
        thread_size = prog["THREAD_SIZE"].value_()
        ranges.append((irq_stack_ptr, irq_stack_ptr + thread_size, StackKind.IRQ))
    except ObjectNotFoundError:
        # IRQ stacks were implemented for arm in d4664b6c987f8 ("ARM: implement
        # IRQ stacks"), starting with v5.18. Prior to that, IRQs were taken on
        # the kernel thread stack.
        pass
    return ranges


def _s390x_stack_ranges(task: Object, cpu: int) -> List[Tuple[int, int, StackKind]]:
    prog = task.prog_

    # Lowcore is a CPU control block which contains, among other thengs, the
    # stack pointers each CPU is configured to use. The stacks pointers are
    # stored with an offset so that they are ready to use by the CPU. But they
    # leave enough room to store a pt_regs and a linkage stack frame at the top.
    lowcore = prog["lowcore_ptr"][cpu]
    thread_size = prog["THREAD_SIZE"].value_()
    stack_init_offset = (
        thread_size
        - sizeof(prog.type("struct stack_frame"))
        - sizeof(prog.type("struct pt_regs"))
    )

    irq = lowcore.async_stack.value_() - stack_init_offset
    restart = lowcore.restart_stack.value_() - stack_init_offset
    ranges = [
        (irq, irq + thread_size, StackKind.IRQ),
        # The remaining arch-specific stacks don't have a better categorization
        # other than "exception"
        (restart, restart + thread_size, StackKind.EXCEPTION),
    ]

    try:
        mcck = lowcore.mcck_stack.value_() - stack_init_offset
        ranges.append((mcck, mcck + thread_size, StackKind.EXCEPTION))
    except AttributeError:
        # mcck_stack is introduced in v5.12 with commit b61b1595124a1 ("s390:
        # add stack for machine check handler").
        pass

    try:
        nodat = lowcore.nodat_stack.value_() - stack_init_offset
    except AttributeError:
        # nodat_stack was panic_stack prior to v4.20, commit ce3dc447493ff
        # ("s390: add support for virtually mapped kernel stacks").
        nodat = lowcore.panic_stack.value_() - stack_init_offset
    ranges.append((nodat, nodat + thread_size, StackKind.EXCEPTION))

    return ranges


def _s390x_classify(_: Object, __: int, frames: List[StackFrame]) -> StackKind:
    # drgn's registers contain the "program status word" which includes the
    # pstate bit, which is set when running in userspace.
    regs = frames[0].registers()
    PSW_MASK_PSTATE = 0x0001000000000000
    pswm = regs.get("pswm")
    if pswm is not None and pswm & PSW_MASK_PSTATE:
        return StackKind.USER
    else:
        return StackKind.UNKNOWN


def _ppc64_stack_ranges(task: Object, cpu: int) -> List[Tuple[int, int, StackKind]]:
    prog = task.prog_
    thread_size = prog["THREAD_SIZE"].value_()

    # These point to the beginning of the stack region
    hardirq = prog["hardirq_ctx"][cpu].value_()
    softirq = prog["softirq_ctx"][cpu].value_()
    ranges = [
        (hardirq, hardirq + thread_size, StackKind.IRQ),
        (softirq, softirq + thread_size, StackKind.SOFTIRQ),
    ]

    try:
        # The PACA (Processor Auxiliary Control Area) defines stack pointers.
        # These are defined differently for "Book 3E" (embedded) and "Book 3S"
        # (server) processors. Drgn tests against Book 3S, so that's what we
        # mainly support.
        paca = prog["paca_ptrs"][cpu]
        if not paca:
            return ranges
    except ObjectNotFoundError:
        # Prior to commit d2e60075a3d44 ("powerpc/64: Use array of paca pointers
        # and allocate pacas individually") in v4.17, this was a struct array,
        # rather than array of pointers to structs.
        paca = prog["paca"][cpu]

    for name in ("emercency_sp", "nmi_emergency_sp", "mc_emergency_sp"):
        sp = int(getattr(paca, name, 0))
        if sp:
            ranges.append((sp - thread_size, sp, StackKind.EXCEPTION))

    return ranges


def _classify_segment(task: Object, cpu: int, frames: List[StackFrame]) -> StackSegment:
    prog = task.prog_
    kind = StackKind.UNKNOWN
    try:
        sp = frames[0].sp
    except LookupError:
        return StackSegment(kind, frames)

    # Architecture-independent: the task stack
    ranges = [
        (
            int(task.stack),
            int(task.stack) + prog["THREAD_SIZE"].value_(),
            StackKind.TASK,
        ),
    ]
    arch = prog.platform.arch if prog.platform else None
    if arch == Architecture.X86_64:
        ranges.extend(_x86_64_stack_ranges(task, cpu))
    elif arch == Architecture.AARCH64:
        ranges.extend(_aarch64_stack_ranges(task, cpu))
    elif arch == Architecture.ARM:
        ranges.extend(_arm_stack_ranges(task, cpu))
    elif arch == Architecture.PPC64:
        ranges.extend(_ppc64_stack_ranges(task, cpu))
    elif arch == Architecture.S390X:
        ranges.extend(_s390x_stack_ranges(task, cpu))

    for start, stop, range_kind in ranges:
        if start <= sp < stop:
            kind = range_kind
            break
    else:
        if arch == Architecture.S390X:
            kind = _s390x_classify(task, cpu, frames)
        elif arch == Architecture.AARCH64:
            kind = _aarch64_classify(task, cpu, frames)

    return StackSegment(kind, frames)


@takes_program_or_default
def kernel_stack_trace(
    prog: Program, task_arg: Union[Object, IntegerLike]
) -> LinuxKernelStack:
    """
    Create a stack trace and return a categorized :class:`LinuxKernelStack`.

    This function creates a stack trace, breaks it into segments, and
    categorizes each segment. The process is best-effort. Not all stack kinds
    can be segmented or recognized, as the process is architecture and version
    specific.

    :param task: a PID or ``struct task_struct *`` to unwind
    :returns: a separated, annotated stack
    """
    if isinstance(task_arg, Object):
        underlying_type = task_arg.type_
        while underlying_type.kind == TypeKind.TYPEDEF:
            underlying_type = underlying_type.type
        if underlying_type.kind == TypeKind.INT:
            pid = operator.index(task_arg)
            task = find_task(prog, pid)
        elif underlying_type.kind == TypeKind.POINTER:
            task = task_arg
        else:
            raise TypeError("expected struct task_struct *, or int (pid)")
    else:
        pid = operator.index(task_arg)
        task = find_task(prog, pid)

    cpu = task_cpu(task)
    trace = task.prog_.stack_trace(task)
    segments: List[StackSegment] = []
    current: List[StackFrame] = []

    for frame_list in _real_frames(trace):
        frame = frame_list[-1]

        if current and frame.interrupted:
            segments.append(_classify_segment(task, cpu, current))
            current = []

        current.extend(frame_list)

    if current:
        segments.append(_classify_segment(task, cpu, current))

    return LinuxKernelStack(task.prog_, segments)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2023, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Interrupts
----------

The ``drgn.helpers.linux.irq`` module provides helpers for working with IRQs
and interrupt descriptors.
"""

from typing import Iterator, List, Optional, Tuple

from drgn import IntegerLike, Object, ObjectNotFoundError, Program, cast
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.cpumask import cpumask_of
from drgn.helpers.linux.mapletree import mt_for_each, mtree_load
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.radixtree import radix_tree_for_each, radix_tree_lookup

__all__ = (
    "for_each_irq_desc",
    "gate_desc_func",
    "irq_desc_action_names",
    "irq_desc_affinity_mask",
    "irq_desc_chip_name",
    "irq_desc_kstat_cpu",
    "irq_to_desc",
)


@takes_program_or_default
def irq_to_desc(prog: Program, irq: IntegerLike) -> Object:
    """
    Get the interrupt descriptor for an IRQ number.

    :param irq: IRQ number.
    :return: ``struct irq_desc *`` (``NULL`` if not found)
    """
    # Since Linux kernel commit 721255b9826b ("genirq: Use a maple tree for
    # interrupt descriptor management") (in v6.5), interrupt descriptors are in
    # a maple tree.
    try:
        sparse_irqs = prog["sparse_irqs"]
    except ObjectNotFoundError:
        pass
    else:
        return cast("struct irq_desc *", mtree_load(sparse_irqs.address_of_(), irq))

    # Before that, they are in a radix tree.
    return cast(
        "struct irq_desc *",
        radix_tree_lookup(prog["irq_desc_tree"].address_of_(), irq),
    )


@takes_program_or_default
def for_each_irq_desc(prog: Program) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all allocated interrupt descriptors.

    :return: Iterator of (IRQ number, ``struct irq_desc *`` object) tuples.
    """
    irq_desc_type = prog.type("struct irq_desc *")

    # Since Linux kernel commit 721255b9826b ("genirq: Use a maple tree for
    # interrupt descriptor management") (in v6.5), interrupt descriptors are in
    # a maple tree.
    try:
        sparse_irqs = prog["sparse_irqs"]
    except ObjectNotFoundError:
        pass
    else:
        for irq, _, entry in mt_for_each(sparse_irqs.address_of_()):
            yield irq, cast(irq_desc_type, entry)
        return

    # Before that, they are in a radix tree.
    for irq, entry in radix_tree_for_each(prog["irq_desc_tree"].address_of_()):
        yield irq, cast(irq_desc_type, entry)


def irq_desc_affinity_mask(desc: Object) -> Object:
    """
    Get the CPU affinity mask for an interrupt descriptor.

    :param desc: ``struct irq_desc *``
    :return: ``struct cpumask *``
    """
    # Since Linux kernel commit aa0813581b8d ("genirq: Provide an IRQ affinity
    # mask in non-SMP configs") (in v6.0), struct irq_common_data::affinity
    # doesn't exist on !SMP. Before that, it exists on !SMP but is
    # uninitialized. So, we need to check for !SMP first.
    if "nr_cpu_ids" not in desc.prog_:
        return cpumask_of(desc.prog_, 0)

    try:
        # affinity is a cpumask_var_t, which is a pointer if
        # CONFIG_CPUMASK_OFFSTACK=y and an array otherwise. The `+ 0` converts
        # it to a pointer in the array case and is a no-op in the pointer case.
        return desc.irq_common_data.affinity + 0
    except AttributeError:
        # Before Linux kernel commit 9df872faa7e1 ("genirq: Move field
        # 'affinity' from irq_data into irq_common_data") (in v4.3), the
        # affinity member is in struct irq_data instead of struct
        # irq_common_data.
        return desc.irq_data.affinity + 0


def irq_desc_action_names(desc: Object) -> List[bytes]:
    """
    Get a list of the names of actions set on an interrupt descriptor.

    Actions without names (e.g., ``chained_action``) are skipped.

    >>> irq_desc_action_names(irq_to_desc(27))
    [b'i2c_designware.0', b'idma64.0']

    :param desc: ``struct irq_desc *``
    """
    action = desc.action.read_()
    names = []
    while action:
        name = action.name.read_()
        if name:
            names.append(name.string_())
        action = action.next.read_()
    return names


def irq_desc_chip_name(desc: Object) -> Optional[bytes]:
    """
    Get the name of the controller chip that manages an interrupt descriptor.

    :param desc: ``struct irq_desc *``
    :return: Chip name, or ``None`` if the chip or name is not set.
    """
    chip = desc.irq_data.chip.read_()
    if chip:
        name = chip.name.read_()
        if name:
            return name.string_()
    return None


def irq_desc_kstat_cpu(desc: Object, cpu: IntegerLike) -> int:
    """
    Get the number of times that an interrupt has fired on a given CPU.

    :param desc: ``struct irq_desc *``
    :param cpu: CPU number.
    """
    kstat_irqs = desc.kstat_irqs.read_()
    if not kstat_irqs:
        return 0
    # Since Linux kernel commit 86d2a2f51fba ("genirq: Convert kstat_irqs to a
    # struct") (in v6.10), kstat_irqs is a structure with the actual counter in
    # a member. Before that, kstat_irqs was the counter itself.
    try:
        cnt = kstat_irqs.cnt
    except AttributeError:
        cnt = kstat_irqs[0]
    return per_cpu(cnt, cpu).value_()


def gate_desc_func(gate: Object) -> Object:
    """
    Get the IDT entry function for a gate on x86.

    >>> gate_desc_func(prog["idt_table"][3])
    (void *)asm_exc_int3+0x0 = 0xffffffff91001240

    :param gate: ``gate_desc`` or ``gate_desc *``
    :return: ``void *``
    """
    return Object(
        gate.prog_,
        "void *",
        (gate.offset_high.value_() << 32)
        | (gate.offset_middle.value_() << 16)
        | gate.offset_low.value_(),
    )



#!/usr/bin/env drgn

import argparse
import os
import sys
import typing
from typing import Any, Callable, Iterator, Optional, Sequence, Union

from drgn import FaultError, Object, Program, TypeKind, cast, container_of
from drgn.helpers.linux.cpumask import for_each_possible_cpu
from drgn.helpers.linux.fs import (
    d_path,
    fget,
    for_each_file,
    for_each_mount,
    inode_path,
    mount_dst,
)
from drgn.helpers.linux.idr import idr_for_each_entry
from drgn.helpers.linux.list import (
    hlist_for_each_entry,
    list_empty,
    list_for_each_entry,
)
from drgn.helpers.linux.mm import for_each_vma
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.pid import find_task, for_each_task
from drgn.helpers.linux.plist import plist_for_each_entry
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry


class warn_on_fault:
    def __init__(self, message: Union[str, Callable[[], str]]) -> None:
        self._message = message

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> bool:
        if exc_type is not None and issubclass(exc_type, FaultError):
            message = (
                self._message if isinstance(self._message, str) else self._message()
            )
            if message:
                print(
                    f"warning: fault while {message}, possibly due to race; results may be incomplete",
                    file=sys.stderr,
                )
            return True
        return False


ignore_fault = warn_on_fault("")


format_args = {
    "dereference": False,
    "symbolize": False,
}


if typing.TYPE_CHECKING:

    class Visitor(typing.Protocol):  # novermin
        def visit_file(self, file: Object) -> Optional[str]: ...

        def visit_inode(self, inode: Object) -> Optional[str]: ...

        def visit_path(self, path: Object) -> Optional[str]: ...


class InodeVisitor:
    def __init__(self, inode: Object) -> None:
        self._inode = inode.read_()

    def visit_file(self, file: Object) -> Optional[str]:
        if file.f_inode != self._inode:
            return None
        return file.format_(**format_args)

    def visit_inode(self, inode: Object) -> Optional[str]:
        if inode != self._inode:
            return None
        return inode.format_(**format_args)

    def visit_path(self, path: Object) -> Optional[str]:
        if path.dentry.d_inode != self._inode:
            return None
        return path.format_(**format_args)


def _format_file_with_path(file: Object) -> str:
    match = file.format_(**format_args)
    with ignore_fault:
        match += " " + os.fsdecode(d_path(file.f_path))
    return match


def _format_inode_with_path(inode: Object) -> str:
    match = inode.format_(**format_args)
    with ignore_fault:
        path = inode_path(inode)
        if path:
            match += " " + os.fsdecode(path)
    return match


def _format_path_with_path(path: Object) -> str:
    match = path.format_(**format_args)
    with ignore_fault:
        match += " " + os.fsdecode(d_path(path))
    return match


class SuperBlockVisitor:
    def __init__(self, sb: Object) -> None:
        self._sb = sb.read_()

    def visit_file(self, file: Object) -> Optional[str]:
        if file.f_inode.i_sb != self._sb:
            return None
        return _format_file_with_path(file)

    def visit_inode(self, inode: Object) -> Optional[str]:
        if inode.i_sb != self._sb:
            return None
        return _format_inode_with_path(inode)

    def visit_path(self, path: Object) -> Optional[str]:
        if path.mnt.mnt_sb != self._sb:
            return None
        return _format_path_with_path(path)


class BtrfsSubvolumeVisitor:
    def __init__(self, root: Object) -> None:
        self._root = root.read_()
        # We compare the super block first to easily filter out non-Btrfs
        # inodes.
        self._sb = self._root.fs_info.sb.read_()

    def visit_file(self, file: Object) -> Optional[str]:
        f_inode = file.f_inode.read_()
        if (
            f_inode.i_sb != self._sb
            or container_of(f_inode, "struct btrfs_inode", "vfs_inode").root
            != self._root
        ):
            return None
        return _format_file_with_path(file)

    def visit_inode(self, inode: Object) -> Optional[str]:
        if (
            inode.i_sb != self._sb
            or container_of(inode, "struct btrfs_inode", "vfs_inode").root != self._root
        ):
            return None
        return _format_inode_with_path(inode)

    def visit_path(self, path: Object) -> Optional[str]:
        inode = path.dentry.d_inode.read_()
        if (
            inode.i_sb != self._sb
            or container_of(inode, "struct btrfs_inode", "vfs_inode").root != self._root
        ):
            return None
        return _format_path_with_path(path)


def super_block_on_bdev(bdev: Object) -> Optional[Object]:
    prog = bdev.prog_

    try:
        btrfs_fs_info_type = prog.type("struct btrfs_fs_info *")
    except LookupError:
        btrfs_fs_info_type = None

    for sb in list_for_each_entry(
        "struct super_block", prog["super_blocks"].address_of_(), "s_list"
    ):
        if btrfs_fs_info_type is not None and sb.s_type.name.string_() == b"btrfs":
            # Btrfs is a special case because of its multi-device support.
            fs_info = cast(btrfs_fs_info_type, sb.s_fs_info)
            for device in list_for_each_entry(
                "struct btrfs_device",
                fs_info.fs_devices.devices.address_of_(),
                "dev_list",
            ):
                if device.bdev == bdev:
                    return sb
        elif sb.s_bdev == bdev:
            return sb

    return None


def visit_tasks(
    prog: Program, visitor: "Visitor", *, check_mounts: bool, check_tasks: bool
) -> None:
    check_mounts = check_mounts and isinstance(visitor, SuperBlockVisitor)
    if check_mounts:
        init_mnt_ns = prog["init_task"].nsproxy.mnt_ns
        checked_mnt_ns = {0}
    with warn_on_fault("iterating tasks"):
        for task in for_each_task(prog):
            cached_task_id = None

            def task_id() -> str:
                nonlocal cached_task_id
                if cached_task_id is None:
                    pid = task.pid.value_()
                    comm = os.fsdecode(task.comm.string_())
                    cached_task_id = f"pid {pid} ({comm})"
                return cached_task_id

            def task_fault_warning() -> str:
                try:
                    return f"checking {task_id()}"
                except FaultError:
                    return "checking task"

            with warn_on_fault(task_fault_warning):
                files: Optional[Object] = task.files.read_()
                fs: Optional[Object] = task.fs.read_()
                mm: Optional[Object] = task.mm.read_()

                # If this task is not the thread group leader, don't bother
                # checking it again unless it has its own context.
                group_leader = task.group_leader.read_()
                if task != group_leader:
                    if files and files == group_leader.files:
                        files = None
                    if fs and fs == group_leader.fs:
                        fs = None
                    if mm and mm == group_leader.mm:
                        mm = None

                if check_mounts:
                    nsproxy = task.nsproxy.read_()
                    if nsproxy:
                        mnt_ns = nsproxy.mnt_ns.read_()
                        if mnt_ns.value_() not in checked_mnt_ns:
                            for mount in for_each_mount(mnt_ns):
                                with ignore_fault:
                                    if mount.mnt.mnt_sb == visitor._sb:  # type: ignore [attr-defined]
                                        if mnt_ns == init_mnt_ns:
                                            mnt_ns_note = ""
                                        else:
                                            mnt_ns_note = f" (mount namespace {mnt_ns.ns.inum.value_()})"
                                        print(
                                            f"mount {os.fsdecode(mount_dst(mount))}{mnt_ns_note} "
                                            f"{mount.format_(**format_args)}"
                                        )

                            checked_mnt_ns.add(mnt_ns.value_())

                if check_tasks:
                    if files:
                        for fd, file in for_each_file(task):
                            with ignore_fault:
                                match = visitor.visit_file(file)
                                if match:
                                    print(f"{task_id()} fd {fd} {match}")

                    if fs:
                        with ignore_fault:
                            match = visitor.visit_path(fs.root.address_of_())
                            if match:
                                print(f"{task_id()} root {match}")
                        with ignore_fault:
                            match = visitor.visit_path(fs.pwd.address_of_())
                            if match:
                                print(f"{task_id()} cwd {match}")

                    if mm:
                        exe_file = mm.exe_file.read_()
                        if exe_file:
                            match = visitor.visit_file(exe_file)
                            if match:
                                print(f"{task_id()} exe {match}")

                        for vma in for_each_vma(mm):
                            with ignore_fault:
                                file = vma.vm_file.read_()
                                if file:
                                    match = visitor.visit_file(file)
                                    if match:
                                        print(
                                            f"{task_id()} vma {hex(vma.vm_start)}-{hex(vma.vm_end)} {match}"
                                        )


def visit_binfmt_misc(prog: Program, visitor: "Visitor") -> None:
    try:
        Node = prog.type("Node", filename="binfmt_misc.c")
    except LookupError:
        # If the Node type doesn't exist, then CONFIG_BINFMT_MISC=n or the
        # binfmt_misc module isn't loaded.
        return
    with warn_on_fault("iterating binfmt_misc instances"):
        for sb in hlist_for_each_entry(
            "struct super_block", prog["bm_fs_type"].fs_supers, "s_instances"
        ):
            # Since Linux kernel commit 21ca59b365c0 ("binfmt_misc: enable
            # sandboxed mounts") (in v6.7), each user namespace can have its
            # own binfmt_misc instance. Before that, there is one global
            # instance.
            user_ns = cast("struct user_namespace *", sb.s_fs_info)
            try:
                binfmt_misc = user_ns.binfmt_misc
            except AttributeError:
                entries = prog.object("entries", filename="binfmt_misc.c")
                have_user_ns = False
            else:
                entries = binfmt_misc.entries
                have_user_ns = True

            for node in list_for_each_entry(Node, entries.address_of_(), "list"):
                with ignore_fault:
                    match = visitor.visit_file(node.interp_file)
                    if match:
                        if have_user_ns and user_ns.level:
                            user_ns_note = (
                                f" (user namespace {user_ns.ns.inum.value_()})"
                            )
                        else:
                            user_ns_note = ""
                        print(
                            f"binfmt_misc{user_ns_note} {os.fsdecode(node.name.string_())} {node.format_(**format_args)} {match}"
                        )


def visit_loop_devices(prog: Program, visitor: "Visitor") -> None:
    try:
        loop_index_idr = prog["loop_index_idr"]
    except KeyError:
        # If loop_index_idr doesn't exist, then CONFIG_BLK_DEV_LOOP=n or the
        # loop module isn't loaded.
        return
    with warn_on_fault("iterating loop devices"):
        for i, lo in idr_for_each_entry(
            loop_index_idr.address_of_(), "struct loop_device"
        ):
            with ignore_fault:
                file = lo.lo_backing_file.read_()
                if file:
                    match = visitor.visit_file(file)
                    if match:
                        print(f"loop device {i} {lo.format_(**format_args)} {match}")


def visit_swap_files(prog: Program, visitor: "Visitor") -> None:
    try:
        swap_active_head = prog["swap_active_head"]
    except KeyError:
        # If swap_active_head doesn't exist, then CONFIG_SWAP=n.
        return
    with warn_on_fault("iterating swap files"):
        for swap_info in plist_for_each_entry(
            "struct swap_info_struct", swap_active_head.address_of_(), "list"
        ):
            with ignore_fault:
                match = visitor.visit_file(swap_info.swap_file)
                if match:
                    print(f"swap file {swap_info.format_(**format_args)} {match}")


# call was moved from struct trace_probe to struct trace_event Linux kernel
# commit 60d53e2c3b75 ("tracing/probe: Split trace_event related data from
# trace_probe") (in v5.4).
def _trace_probe_call(tp: Object) -> Object:
    try:
        event = tp.event
    except AttributeError:
        return tp.call
    return event.call


def trace_probe_group_name(tp: Object) -> str:
    return os.fsdecode(_trace_probe_call(tp).member_("class").system.string_())


def trace_probe_name(tp: Object) -> str:
    prog = tp.prog_
    call = _trace_probe_call(tp).read_()

    # TRACE_EVENT_FL_CUSTOM was added in Linux kernel commit 3a73333fb370
    # ("tracing: Add TRACE_CUSTOM_EVENT() macro") (in v5.18).
    try:
        TRACE_EVENT_FL_CUSTOM = prog["TRACE_EVENT_FL_CUSTOM"]
    except KeyError:
        pass
    else:
        if call.flags & TRACE_EVENT_FL_CUSTOM:
            return os.fsdecode(call.name.string_())

    if call.flags & prog["TRACE_EVENT_FL_TRACEPOINT"]:
        tracepoint = call.tp.read_()
        return os.fsdecode(tracepoint.name.string_()) if tracepoint else ""
    else:
        return os.fsdecode(call.name.string_())


def visit_uprobes(prog: Program, visitor: "Visitor") -> None:
    try:
        uprobes_tree = prog["uprobes_tree"]
    except KeyError:
        # If uprobes_tree doesn't exist, then CONFIG_UPROBES=n.
        return
    try:
        uprobe_dispatcher = prog["uprobe_dispatcher"]
    except KeyError:
        # uprobe_dispatcher only exists if CONFIG_UPROBE_EVENTS=y, which is
        # theoretically separate from CONFIG_UPROBES, although as of Linux 6.8
        # they will always be the same.
        uprobe_dispatcher = None
    with warn_on_fault("iterating uprobes"):
        for uprobe in rbtree_inorder_for_each_entry(
            "struct uprobe", uprobes_tree.address_of_(), "rb_node"
        ):
            try:
                match = visitor.visit_inode(uprobe.inode)
            except FaultError:
                continue
            if not match:
                continue
            found_consumer = False
            with warn_on_fault("iterating uprobe consumers"):
                consumer_list = uprobe.consumers
                # Since Linux kernel commit cc01bd044e6a ("uprobes: travers
                # uprobe's consumer list locklessly under SRCU protection") (in
                # v6.12), struct uprobe::consumers is a struct list_head.
                # Before that, it was a hand-rolled singly-linked list.
                if consumer_list.type_.kind == TypeKind.STRUCT:
                    consumers = list_for_each_entry(
                        "struct uprobe_consumer",
                        consumer_list.address_of_(),
                        "cons_node",
                    )
                else:

                    def for_each_consumer(first: Object) -> Iterator[Object]:
                        consumer = first.read_()
                        while consumer:
                            yield consumer
                            consumer = consumer.next.read_()

                    consumers = for_each_consumer(consumer_list)

                for consumer in consumers:
                    handler = consumer.handler.read_()
                    if handler == uprobe_dispatcher:
                        tu = container_of(consumer, "struct trace_uprobe", "consumer")
                        # uprobe events created through tracefs are in a list
                        # anchored on devent.list since Linux kernel commit
                        # 0597c49c69d5 ("tracing/uprobes: Use dyn_event
                        # framework for uprobe events") (in v5.0) and list
                        # before that.
                        try:
                            event_list = tu.devent.list
                        except AttributeError:
                            event_list = tu.list
                        if list_empty(event_list.address_of_()):
                            found_perf_event = False
                            with ignore_fault:
                                call = _trace_probe_call(tu.tp)
                                # uprobes created with perf_event_open have a
                                # struct perf_event in call.perf_events, which
                                # only exists if CONFIG_PERF_EVENTS=y.
                                try:
                                    perf_events = call.perf_events
                                except AttributeError:
                                    pass
                                else:
                                    for cpu in for_each_possible_cpu(prog):
                                        for perf_event in hlist_for_each_entry(
                                            "struct perf_event",
                                            per_cpu_ptr(perf_events, cpu),
                                            "hlist_entry",
                                        ):
                                            owner = perf_event.owner.read_()
                                            if owner:
                                                owner_pid = owner.pid.value_()
                                                owner_comm = os.fsdecode(
                                                    owner.comm.string_()
                                                )
                                                print(
                                                    f"perf uprobe (owned by pid {owner_pid} ({owner_comm})) {perf_event.format_(**format_args)} {match}"
                                                )
                                            else:
                                                print(
                                                    f"perf uprobe (no owner) {perf_event.format_(**format_args)} {match}"
                                                )
                                            found_perf_event = True
                            if not found_perf_event:
                                print(
                                    f"unknown trace uprobe {tu.format_(**format_args)} {match}"
                                )
                        else:
                            c = "r" if tu.consumer.ret_handler else "p"
                            group_name = trace_probe_group_name(tu.tp)
                            event_name = trace_probe_name(tu.tp)
                            print(
                                f"uprobe event {c}:{group_name}/{event_name} {tu.format_(**format_args)} {match}"
                            )
                    else:
                        print(
                            f"unknown uprobe consumer {consumer.format_(**format_args)}"
                        )
                    found_consumer = True
            if not found_consumer:
                print(f"unknown uprobe {uprobe.format_(**format_args)} {match}")


def hexint(x: str) -> int:
    return int(x, 16)


def main(prog: Program, argv: Sequence[str]) -> None:
    parser = argparse.ArgumentParser(
        description="find what is referencing a filesystem object"
    )

    parser.add_argument(
        "-L",
        "--dereference",
        action="store_true",
        help="if the given path is a symbolic link, follow it",
    )

    object_group = parser.add_argument_group(
        title="filesystem object selection"
    ).add_mutually_exclusive_group(required=True)
    object_group.add_argument(
        "--inode", metavar="PATH", help="find references to the inode at the given path"
    )
    object_group.add_argument(
        "--inode-pointer",
        metavar="ADDRESS",
        type=hexint,
        help="find references to the given struct inode pointer",
    )
    object_group.add_argument(
        "--super-block",
        metavar="PATH",
        help="find references to the filesystem (super block) containing the given path",
    )
    object_group.add_argument(
        "--super-block-on-block-device",
        metavar="PATH",
        help="find references to the filesystem (super block) on the block device at the given path",
    )
    object_group.add_argument(
        "--super-block-pointer",
        metavar="ADDRESS",
        type=hexint,
        help="find references to the given struct super_block pointer",
    )
    object_group.add_argument(
        "--btrfs-subvolume",
        metavar="PATH",
        help="find references to the Btrfs subvolume containing the given path",
    )
    object_group.add_argument(
        "--btrfs-subvolume-pointer",
        metavar="ADDRESS",
        type=hexint,
        help="find references to the given struct btrfs_root pointer",
    )

    CHECKS = [
        "binfmt_misc",
        "loop",
        "mounts",
        "swap",
        "tasks",
        "uprobes",
    ]
    check_group = parser.add_argument_group(
        title="check selection"
    ).add_mutually_exclusive_group()
    check_group.add_argument(
        "--check",
        choices=CHECKS,
        action="append",
        help="only check for references from the given source; may be given multiple times (default: all)",
    )
    check_group.add_argument(
        "--no-check",
        choices=CHECKS,
        action="append",
        help="don't check for references from the given source; may be given multiple times",
    )

    args = parser.parse_args(argv)

    visitor: "Visitor"
    if args.inode is not None:
        fd = os.open(args.inode, os.O_PATH | (0 if args.dereference else os.O_NOFOLLOW))
        try:
            visitor = InodeVisitor(fget(find_task(prog, os.getpid()), fd).f_inode)
        finally:
            os.close(fd)
    elif args.inode_pointer is not None:
        visitor = InodeVisitor(Object(prog, "struct inode *", args.inode_pointer))
    elif args.super_block is not None:
        fd = os.open(
            args.super_block, os.O_PATH | (0 if args.dereference else os.O_NOFOLLOW)
        )
        try:
            visitor = SuperBlockVisitor(
                fget(find_task(prog, os.getpid()), fd).f_inode.i_sb
            )
        finally:
            os.close(fd)
    elif args.super_block_on_block_device:
        fd = os.open(
            args.super_block_on_block_device,
            # O_PATH doesn't look up the bdev inode, so we can't use it here.
            os.O_RDONLY | (0 if args.dereference else os.O_NOFOLLOW),
        )
        try:
            file = fget(find_task(prog, os.getpid()), fd)
            if file.f_op != prog["def_blk_fops"].address_of_():
                sys.exit(
                    f"{args.super_block_on_block_device} is not a block device. Do you need --dereference?"
                )
            sb = super_block_on_bdev(
                container_of(
                    file.f_mapping.host, "struct bdev_inode", "vfs_inode"
                ).bdev.address_of_()
            )
            if sb is None:
                sys.exit(f"no filesystem found on {args.super_block_on_block_device}")
            visitor = SuperBlockVisitor(sb)
        finally:
            os.close(fd)
    elif args.super_block_pointer is not None:
        visitor = SuperBlockVisitor(
            Object(prog, "struct super_block *", args.super_block_pointer)
        )
    elif args.btrfs_subvolume is not None:
        fd = os.open(
            args.btrfs_subvolume, os.O_PATH | (0 if args.dereference else os.O_NOFOLLOW)
        )
        try:
            inode = fget(find_task(prog, os.getpid()), fd).f_inode.read_()
            # It'd be better to check the filesystem type directly from
            # userspace using fstatfs(2), but Python doesn't provide
            # {,f}statfs() (and os.statvfs() doesn't include f_type).
            if inode.i_sb.s_type.name.string_() != b"btrfs":
                sys.exit(f"{args.btrfs_subvolume} is not on Btrfs")
            visitor = BtrfsSubvolumeVisitor(
                container_of(inode, "struct btrfs_inode", "vfs_inode").root
            )
        finally:
            os.close(fd)
    elif args.btrfs_subvolume_pointer is not None:
        visitor = BtrfsSubvolumeVisitor(
            Object(prog, "struct btrfs_root *", args.btrfs_subvolume_pointer)
        )
    else:
        assert False

    if args.check:
        enabled_checks = set(args.check)
    else:
        enabled_checks = set(CHECKS)
        if args.no_check:
            enabled_checks -= set(args.no_check)

    if "mounts" in enabled_checks or "tasks" in enabled_checks:
        visit_tasks(
            prog,
            visitor,
            check_mounts="mounts" in enabled_checks,
            check_tasks="tasks" in enabled_checks,
        )

    if "binfmt_misc" in enabled_checks:
        visit_binfmt_misc(prog, visitor)

    if "loop" in enabled_checks:
        visit_loop_devices(prog, visitor)

    if "swap" in enabled_checks:
        visit_swap_files(prog, visitor)

    if "uprobes" in enabled_checks:
        visit_uprobes(prog, visitor)


if __name__ == "__main__":
    prog: Program
    main(prog, sys.argv[1:])



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import drgn
from drgn import NULL, Object, cast, container_of, execscript, offsetof, reinterpret, sizeof
from drgn.helpers.common import *
from drgn.helpers.linux import *
import collections
import sys

"""Btrfs Tree Mod Log rewind simulator"""

# you can get a tree mod log from fs_info.tree_mod_log

# search for tree mod log entries for the given offset and minimum sequence
# returns a sorted list of matching entries.
def tree_mod_log_search(tml, start, min_seq):
    es = collections.defaultdict(list)
    for e in rbtree_inorder_for_each_entry("struct tree_mod_elem", tml, "node"):
        es[int(e.logical)].append((int(e.seq), e))
    return [p[1] for p in sorted(es[start]) if p[0] >= min_seq]

# apply the tree mod log entries returned by tree_mod_log_search
# in reverse to a model of a blank extent_buffer. Pay extra attention
# to a particular slot.
def tree_mod_log_rewind(tmes, my_slot):
    eb_rewind = {}
    for tme in reversed(tmes):
        print(tme)
        op = int(tme.op)
        slot = int(tme.slot)
        # replace
        if op == 0:
            if slot == my_slot:
                print(f"writing {tme} into {my_slot}!")
            eb_rewind[slot] = (tme.blockptr, tme.generation, tme.key)
        # add
        if op == 1:
            if slot == my_slot:
                print(f"nuking {my_slot}!")
            del(eb_rewind[slot])
        # remove
        if op in [2,3,4]:
            if slot == my_slot:
                print(f"writing {tme} into {my_slot}!")
            eb_rewind[slot] = (int(tme.blockptr), int(tme.generation), tme.key)
        # move
        if op == 5:
            src_slot = int(tme.move.dst_slot)
            nr = int(tme.move.nr_items)
            off = 0
            for src in range(src_slot, src_slot+nr):
                if src in eb_rewind:
                    if slot + off == my_slot:
                        print(f"moving {eb_rewind[src]} into {my_slot}!")
                    eb_rewind[slot + off] = eb_rewind[src]
                else:
                    if slot + off == my_slot:
                        print(f"moving garbage into {my_slot}!")
                    eb_rewind[slot + off] = (0,0,(0,0,0))
                off += 1
    return eb_rewind

# compare the slots of a real eb and a rewound eb
# parsed_eb is the output of 'parse_extent_buffer' in btrfs.py
def diff_eb_rewind(parsed_eb, eb_rewind):
    ebptrs = parsed_eb.ptrs
    mismatch = False
    for i, ptr in enumerate(ebptrs):
        if i in eb_rewind:
            if ptr.blockptr != eb_rewind[i][0]:
                mismatch = True
                print(f"EB {i}: {ptr.blockptr} EB_REWIND {i}: {eb_rewind[i][0]}")
        elif ptr.blockptr != 0 and ptr.blockptr < 1 << 41:
            mismatch = True
            print(f"EB ONLY {i}: {ptr.blockptr}")
    return mismatch



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Panic
-----

The ``drgn.helpers.linux.panic`` module provides helpers for getting kernel
panic information.
"""

import re
from typing import Optional

from drgn import NULL, Object, ObjectNotFoundError, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.printk import get_dmesg

__all__ = (
    "panic_message",
    "panic_task",
    "tainted",
)


@takes_program_or_default
def panic_task(prog: Program) -> Object:
    """
    Return the task that panicked.

    This is equivalent to:

    .. code-block:: python3

        from drgn import NULL


        try:
            task = prog.crashed_thread().object
        except ValueError:
            task = NULL(prog, "struct task_struct *")

    :return: ``struct task_struct *`` (``NULL`` if the kernel has not panicked)
    """
    try:
        return prog.crashed_thread().object
    except ValueError:
        return NULL(prog, "struct task_struct *")


# Patterns matching a panic, ranked from most preferred to least preferred.
_PANIC_PATTERNS = (
    rb"Kernel panic -.*",
    # Various architecture faults.
    rb"BUG: (?:kernel NULL pointer dereference|unable to handle).*|Unable to handle kernel.*",
    # BUG_ON() and co.
    rb"[Kk]ernel BUG at.*",
    rb"Oops: .*",
)

_PANIC_PATTERN = b"|".join([b"(%s)" % pattern for pattern in _PANIC_PATTERNS])


def _panic_message(dmesg: bytes) -> Optional[bytes]:
    best_match = None
    best_index = None
    for match in re.finditer(_PANIC_PATTERN, dmesg):
        if best_index is None or match.lastindex < best_index:
            best_match = match.group()
            best_index = match.lastindex
    return best_match


@takes_program_or_default
def panic_message(prog: Program) -> Optional[bytes]:
    """
    Get the kernel message logged during a panic.

    >>> panic_message()
    b'Oops: Oops: 0002 [#1] SMP NOPTI'

    Note that this returns :class:`bytes`. The recommended way to get a
    :class:`str` is with :meth:`bytes.decode(errors="replace") <bytes.decode>`:

    .. code-block:: python3

        message = panic_message()
        if message is not None:
            message = message.decode(errors="replace")

    :return: Message, or ``None`` if the kernel has not panicked or no message
        was found
    """
    return _panic_message(get_dmesg(prog, timestamps=False))


@takes_program_or_default
def tainted(prog: Program) -> str:
    """
    Get the kernel's tainted flags as a string.

    If the kernel is not tainted, then this returns an empty string:

    >>> tainted()
    ''

    Otherwise, it returns the flags as letters:

    >>> tainted()
    'G           O       '

    See the `kernel documentation
    <https://docs.kernel.org/admin-guide/tainted-kernels.html>`_ for an
    explanation of the flags.
    """
    mask = prog["tainted_mask"].value_()
    if not mask:
        return ""

    parts = []
    try:
        taint_flags = prog["taint_flags"]
    except ObjectNotFoundError:
        # Before Linux kernel commit 7fd8329ba502 ("taint/module: Clean up
        # global and module taint flags handling") (in v4.10), the array and
        # members had different names and the bit number was a member.
        for t in prog["tnts"]:
            parts.append(chr(t.true if (mask & (1 << t.bit.value_())) else t.false))  # type: ignore[arg-type]  # python/typeshed#13494
    else:
        for i, t in enumerate(taint_flags):
            parts.append(chr(t.c_true if (mask & (1 << i)) else t.c_false))  # type: ignore[arg-type]  # python/typeshed#13494
    return "".join(parts)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

# ebpf-related commands.

import argparse
from datetime import datetime
from typing import Any, List, Sequence

from drgn import Program
from drgn.commands import argument, drgn_argument
from drgn.commands._crash.common import CrashDrgnCodeBuilder, crash_command
from drgn.helpers.common.format import CellFormat, escape_ascii_string, print_table
from drgn.helpers.linux.bpf import (
    bpf_map_for_each,
    bpf_prog_by_id,
    bpf_prog_for_each,
    bpf_prog_used_maps,
)
from drgn.helpers.linux.timekeeping import (
    ktime_get_coarse_boottime_ns,
    ktime_get_coarse_real_ns,
)
from drgn.helpers.linux.user import kuid_val


@crash_command(
    description="display loaded eBPF programs and maps",
    arguments=(
        argument(
            "-p",
            dest="prog_id",
            type=int,
            help="display additional information for the specified BPF program ID",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_bpf(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.bpf",
            "bpf_map_for_each",
            "bpf_prog_for_each",
            "bpf_prog_used_maps",
        )
        code.append(
            """\
for bpf_prog in bpf_prog_for_each(prog):
    aux = bpf_prog.aux
    prog_id = aux.id
    prog_type = bpf_prog.type
    tag = bpf_prog.tag
    for used_map in bpf_prog_used_maps(bpf_prog):
        map_id = used_map.id

for bpf_map in bpf_map_for_each(prog):
    map_id = bpf_map.id
    map_type = bpf_map.map_type
    try:
        map_flags = bpf_map.map_flags
    except AttributeError:
        # map_flags is only available since Linux 4.6.
        pass
    """
        )
        code.print()
        return
    prog_rows: List[Sequence[Any]] = [
        (
            CellFormat("ID", "^"),
            CellFormat("BPF_PROG", "^"),
            CellFormat("BPF_PROG_AUX", "^"),
            CellFormat("BPF_PROG_TYPE", "^"),
            CellFormat("TAG", "^"),
            CellFormat("USED_MAPS", "^"),
        )
    ]

    if args.prog_id is not None:

        bpf_prog = bpf_prog_by_id(prog, args.prog_id)
        if not bpf_prog:
            print(f"invalid BPF program ID: {args.prog_id}")
            return

        aux = bpf_prog.aux
        prog_id = aux.id.value_()
        prog_type_name = bpf_prog.type.format_(type_name=False).split("BPF_PROG_TYPE_")[
            -1
        ]
        PAGE_SIZE = prog["PAGE_SIZE"].value_()
        INSN_SIZE = prog.type("struct bpf_insn").size

        tag = bpf_prog.tag
        prog_tag = "".join(f"{b.value_():02x}" for b in tag)

        used_maps = []
        for map in bpf_prog_used_maps(bpf_prog):
            used_maps.append(map.id.value_())
        used_maps_str = ",".join(str(m) for m in used_maps)

        prog_rows.append(
            (
                prog_id,
                CellFormat(bpf_prog.value_(), "^x"),
                CellFormat(aux.value_(), "^x"),
                CellFormat(prog_type_name, "^"),
                CellFormat(prog_tag, "^"),
                CellFormat(used_maps_str, "^"),
            )
        )
        print_table(prog_rows)

        print(
            f"     XLATED: {bpf_prog.len.value_() * INSN_SIZE}  JITED: {bpf_prog.jited_len.value_()}  MEMLOCK: {bpf_prog.pages.value_() * PAGE_SIZE}"
        )

        # load_time and name were added in Linux kernel commit cb4d2b3f03d8 ("bpf:
        # Add name, load_time, uid and map_ids to bpf_prog_info") (in v4.15).
        # Before that, default to "(unknown)".
        try:
            load_time_ns = aux.load_time.value_()
        except AttributeError:
            load_time_str = "(unknown)"
        else:
            current_boottime_ns = ktime_get_coarse_boottime_ns(prog).value_()
            current_realtime_ns = ktime_get_coarse_real_ns(prog).value_()

            elapsed_ns = current_boottime_ns - load_time_ns
            actual_load_time_ns = current_realtime_ns - elapsed_ns
            actual_load_time = datetime.fromtimestamp(actual_load_time_ns / 1e9)
            load_time_str = actual_load_time.strftime("%a %b %d %H:%M:%S %Y")

        print(f"     LOAD_TIME: {load_time_str}")

        gpl_compat = "yes" if bpf_prog.gpl_compatible else "no"
        uid = kuid_val(aux.user.uid)
        try:
            prog_name = escape_ascii_string(aux.name.string_(), escape_backslash=True)
            if not prog_name:
                prog_name = "(unused)"
        except AttributeError:
            prog_name = "(unknown)"

        print(f"     GPL_COMPATIBLE: {gpl_compat}  NAME: {prog_name}  UID: {uid}")

        return

    for bpf_prog in bpf_prog_for_each(prog):
        prog_id = bpf_prog.aux.id.value_()
        prog_type_name = bpf_prog.type.format_(type_name=False).split("BPF_PROG_TYPE_")[
            -1
        ]

        tag = bpf_prog.tag
        prog_tag = "".join(f"{b.value_():02x}" for b in tag)

        used_maps = []
        for map in bpf_prog_used_maps(bpf_prog):
            used_maps.append(map.id.value_())
        used_maps_str = ",".join(str(m) for m in used_maps)

        prog_rows.append(
            (
                prog_id,
                CellFormat(bpf_prog.value_(), "^x"),
                CellFormat(bpf_prog.aux.value_(), "^x"),
                CellFormat(prog_type_name, "^"),
                CellFormat(prog_tag, "^"),
                CellFormat(used_maps_str, "^"),
            )
        )

    print_table(prog_rows)
    print()

    map_rows: List[Sequence[Any]] = [
        (
            CellFormat("ID", "^"),
            CellFormat("BPF_MAP", "^"),
            CellFormat("BPF_MAP_TYPE", "^"),
            CellFormat("MAP_FLAGS", "^"),
        )
    ]

    for bpf_map in bpf_map_for_each(prog):
        map_id = bpf_map.id.value_()

        map_type_name = bpf_map.map_type.format_(type_name=False).split(
            "BPF_MAP_TYPE_"
        )[-1]

        # The 'map_flags' field was added in Linux kernel commit 6c9059817432
        # ("bpf: pre-allocate hash map elements") in version 4.6. Older kernels
        # may not have this field, so we catch LookupError and default to 0.
        try:
            map_flags = f"{bpf_map.map_flags.value_():08x}"
        except AttributeError:
            map_flags = "00000000"

        map_rows.append(
            (
                map_id,
                CellFormat(bpf_map.value_(), "^x"),
                CellFormat(map_type_name, "^"),
                CellFormat(map_flags, "^"),
            )
        )

    print_table(map_rows)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""kmem command."""

import argparse
import functools
import operator
import sys
from typing import (
    AbstractSet,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from drgn import (
    NULL,
    FaultError,
    Object,
    Program,
    ProgramFlags,
    RelocatableModule,
    TypeKind,
    cast,
    sizeof,
)
from drgn.commands import (
    CommandArgumentError,
    _repr_black,
    argument,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash._sys import _print_sys
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _parse_members,
    _sanitize_member_name,
    crash_command,
)
from drgn.helpers import ValidationError
from drgn.helpers.common.format import (
    CellFormat,
    RowOptions,
    _print_table_row,
    double_quote_ascii_string,
    escape_ascii_string,
    number_in_binary_units,
    print_table,
)
from drgn.helpers.common.memory import IdentifiedSymbol, identify_address_all
from drgn.helpers.common.type import typeof_member
from drgn.helpers.linux.block import nr_blockdev_pages
from drgn.helpers.linux.common import (
    IdentifiedPage,
    IdentifiedSlabObject,
    IdentifiedTaskStack,
    IdentifiedTaskStruct,
    IdentifiedVmap,
)
from drgn.helpers.linux.cpumask import for_each_possible_cpu
from drgn.helpers.linux.device import dev_name
from drgn.helpers.linux.hugetlb import (
    for_each_hstate,
    huge_page_size,
    hugetlb_total_usage,
)
from drgn.helpers.linux.list import (
    validate_list_count_nodes,
    validate_list_for_each_entry,
)
from drgn.helpers.linux.mm import (
    PFN_PHYS,
    decode_memory_block_state,
    decode_page_flags_value,
    follow_pfn,
    for_each_memory_block,
    for_each_page,
    for_each_valid_pfn_and_page,
    for_each_vmap_area,
    memory_block_size_bytes,
    page_flags,
    page_index,
    page_to_pfn,
    page_to_virt,
    pfn_to_page,
    totalram_pages,
    vm_commit_limit,
    vm_memory_committed,
)
from drgn.helpers.linux.mmzone import (
    NODE_DATA,
    decode_section_flags,
    for_each_online_pgdat,
    for_each_present_section,
    high_wmark_pages,
    low_wmark_pages,
    min_wmark_pages,
    section_decode_mem_map,
    section_mem_map_addr,
    section_nr_to_pfn,
)
from drgn.helpers.linux.nodemask import for_each_online_node
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.slab import (
    find_slab_cache,
    for_each_slab_cache,
    slab_cache_order,
    slab_cache_usage,
    slab_total_usage,
)
from drgn.helpers.linux.swap import swap_total_usage, total_swapcache_pages
from drgn.helpers.linux.vmstat import (
    global_node_page_state,
    global_numa_event_state,
    global_vm_event_state,
    global_zone_page_state,
    nr_free_pages,
    zone_page_state,
)


def _kmem_free(prog: Program, drgn_arg: bool, show_pages: bool = False) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.mm", "PFN_PHYS", "pfn_to_page")
        code.add_from_import("drgn.helpers.linux.mmzone", "for_each_online_pgdat")
        code.add_from_import(
            "drgn.helpers.linux.vmstat", "nr_free_pages", "zone_page_state"
        )
        with code.begin_block(
            """\
actual_free_pages = 0
for pgdat in for_each_online_pgdat():
    for zone in pgdat.node_zones:
        name = zone.name
        size = zone.spanned_pages
        if size == 0:
            continue
        free = zone_page_state(zone, prog["NR_FREE_PAGES"])
        start_pfn = zone.zone_start_pfn.read_()
        mem_map = pfn_to_page(start_pfn)
        start_paddr = PFN_PHYS(start_pfn)

        for order, free_area in enumerate(zone.free_area):
            block_size = prog["PAGE_SIZE"] << order
            for migrate_type, free_list in enumerate(free_area.free_list):
"""
        ), code.begin_retry_loop_if_live(100):
            if show_pages:
                code.add_from_import(
                    "drgn.helpers.linux.list", "validate_list_for_each_entry"
                )
                code.append(
                    """\
num_blocks = 0
for page in validate_list_for_each_entry(
    "struct page", free_list.address_of_(), "lru"
):
    num_blocks += 1
"""
                )
            else:
                code.add_from_import(
                    "drgn.helpers.linux.list", "validate_list_count_nodes"
                )
                code.append(
                    "num_blocks = validate_list_count_nodes(free_list.address_of_())\n"
                )
        code.append(
            """\
                num_pages = num_blocks << order
                actual_free_pages += num_pages

expected_free_pages = nr_free_pages()
"""
        )
        code.print()
        return

    migrate_types = {
        enumerator.value: enumerator.name[len("MIGRATE_") :]
        # enum migratetype was anonymous until Linux kernel commit a6ffdc07847e
        # ("mm: use is_migrate_highatomic() to simplify the code") (in v4.12).
        for enumerator in prog["MIGRATE_UNMOVABLE"].type_.enumerators  # type: ignore[union-attr]
        if enumerator.name.startswith("MIGRATE_")
    }

    pgdats = list(for_each_online_pgdat(prog))
    if len(pgdats) > 1:
        node_header: Sequence[Any] = (CellFormat("NODE", ">"),)
    else:
        node_header = ()

    free_list_header = [
        CellFormat("ORDER", "^"),
        CellFormat("SIZE", ">"),
        "MIGRATE",
        CellFormat("FREE_LIST", "^"),
    ]
    if not show_pages:
        free_list_header.append(CellFormat("BLOCKS", ">"))
        free_list_header.append(CellFormat("PAGES", ">"))

    rows: List[Sequence[Any]] = []
    first = True
    total_free_pages = 0
    for pgdat in pgdats:
        for i, zone in enumerate(pgdat.node_zones):
            if len(pgdats) > 1:
                node_row: Sequence[Any] = (pgdat.node_id.value_(),)
            else:
                node_row = ()
            size = zone.spanned_pages.value_()

            if first:
                first = False
            else:
                rows.append(())
            rows.append(
                RowOptions(
                    (
                        *node_header,
                        CellFormat("ZONE", "^"),
                        "NAME",
                        CellFormat("SIZE", ">"),
                        CellFormat("FREE", ">"),
                        CellFormat("MEM_MAP", "^"),
                        "START_PADDR",
                        "START_PFN",
                    ),
                    group=1,
                )
            )

            if size == 0:
                free = 0
                mem_map_cell = start_paddr_cell = start_pfn_cell = CellFormat(0, "^")
            else:
                free = zone_page_state(zone, prog["NR_FREE_PAGES"])
                start_pfn = zone.zone_start_pfn.read_()
                mem_map_cell = CellFormat(pfn_to_page(start_pfn).value_(), "^x")
                start_paddr_cell = CellFormat(PFN_PHYS(start_pfn).value_(), "^x")
                start_pfn_cell = CellFormat(start_pfn.value_(), "^")
            rows.append(
                RowOptions(
                    (
                        *node_row,
                        CellFormat(i, "^"),
                        escape_ascii_string(zone.name.string_(), escape_backslash=True),
                        size,
                        free,
                        mem_map_cell,
                        start_paddr_cell,
                        start_pfn_cell,
                    ),
                    group=1,
                )
            )

            if size == 0:
                continue

            if not show_pages:
                rows.append(free_list_header)
            for order, free_area in enumerate(zone.free_area):
                size = (prog["PAGE_SIZE"].value_() << order) // 1024
                size_cell = CellFormat(f"{size}k", ">")
                for migrate_type, free_list in enumerate(free_area.free_list):
                    num_blocks: Any = "[CORRUPTED]"
                    num_pages: Any = ""
                    # Walking free lists is racy. Retry a limited number of
                    # times on live kernels.
                    for _ in range(10):
                        try:
                            if show_pages:
                                pages = list(
                                    validate_list_for_each_entry(
                                        "struct page", free_list.address_of_(), "lru"
                                    )
                                )
                                num_blocks = len(pages)
                            else:
                                num_blocks = validate_list_count_nodes(
                                    free_list.address_of_()
                                )
                        except (FaultError, ValidationError):
                            if not (prog.flags & ProgramFlags.IS_LIVE):
                                break
                        else:
                            num_pages = num_blocks << order
                            total_free_pages += num_pages
                            break
                    if show_pages:
                        rows.append(free_list_header)
                    row = [
                        CellFormat(order, "^"),
                        size_cell,
                        migrate_types[migrate_type],
                        CellFormat(free_list.address_, "^x"),
                    ]
                    if not show_pages:
                        row.append(num_blocks)
                        row.append(num_pages)
                    rows.append(row)

                    if show_pages:
                        for page in pages:
                            rows.append(
                                RowOptions((CellFormat(page.value_(), "x"),), group=2)
                            )
    print_table(rows)

    free_pages = nr_free_pages(prog)
    if free_pages == total_free_pages:
        verified = "verified"
    else:
        verified = f"found {total_free_pages}"
    print(f"\nnr_free_pages: {free_pages}  ({verified})")


def _kmem_info(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.block", "nr_blockdev_pages")
        code.add_from_import("drgn.helpers.linux.hugetlb", "hugetlb_total_usage")
        code.add_from_import(
            "drgn.helpers.linux.mm",
            "totalram_pages",
            "vm_commit_limit",
            "vm_memory_committed",
        )
        code.add_from_import("drgn.helpers.linux.slab", "slab_total_usage")
        code.add_from_import(
            "drgn.helpers.linux.vmstat", "global_node_page_state", "nr_free_pages"
        )
        code.add_from_import(
            "drgn.helpers.linux.swap", "swap_total_usage", "total_swapcache_pages"
        )
        code.append(
            """\
total_mem = totalram_pages()
free = nr_free_pages()
used = total_mem - free
buffers = nr_blockdev_pages()
cached = (
    global_node_page_state(prog["NR_FILE_PAGES"]) - total_swapcache_pages() - buffers
)
slab_usage = slab_total_usage()

hugetlb_usage = hugetlb_total_usage()

swap_usage = swap_total_usage()

commit_limit = vm_commit_limit()
committed = vm_memory_committed()
"""
        )
        code.print()
        return

    rows: List[Sequence[Any]] = [
        (
            "",
            CellFormat("PAGES", ">"),
            CellFormat("TOTAL", ">"),
            CellFormat("PERCENTAGE", "^"),
        )
    ]

    page_size = prog["PAGE_SIZE"].value_()

    def binary_units_cell(n: int) -> CellFormat:
        return CellFormat(
            "0" if n == 0 else number_in_binary_units(n * page_size) + "B", ">"
        )

    total_mem = totalram_pages(prog)
    rows.append(
        (
            CellFormat("TOTAL MEM", ">"),
            total_mem,
            binary_units_cell(total_mem),
            CellFormat("----", "^"),
        )
    )

    free = nr_free_pages(prog)
    buffers = nr_blockdev_pages(prog)
    cached = (
        global_node_page_state(prog["NR_FILE_PAGES"])
        - total_swapcache_pages(prog)
        - buffers
    )

    for label, pages in (
        ("FREE", free),
        ("USED", total_mem - free),
        ("BUFFERS", buffers),
        ("CACHED", cached),
        ("SLAB", slab_total_usage(prog).total_pages),
    ):
        rows.append(
            (
                CellFormat(label, ">"),
                pages,
                binary_units_cell(pages),
                CellFormat(f"{pages / total_mem:.0%} of TOTAL MEM", ">"),
            ),
        )

    rows.append(())

    hugetlb_usage = hugetlb_total_usage(prog)
    rows.append(
        (
            CellFormat("TOTAL HUGE", ">"),
            hugetlb_usage.pages,
            binary_units_cell(hugetlb_usage.pages),
            CellFormat("----", "^"),
        )
    )
    rows.append(
        (
            CellFormat("HUGE FREE", ">"),
            hugetlb_usage.free_pages,
            binary_units_cell(hugetlb_usage.free_pages),
            CellFormat(
                f"{hugetlb_usage.free_pages / hugetlb_usage.pages if hugetlb_usage.pages else 0:.0%} of TOTAL HUGE",
                ">",
            ),
        )
    )

    rows.append(())

    swap_usage = swap_total_usage(prog)
    rows.append(
        (
            CellFormat("TOTAL SWAP", ">"),
            swap_usage.pages,
            binary_units_cell(swap_usage.pages),
            CellFormat("----", "^"),
        )
    )
    for label, pages in (
        ("SWAP USED", swap_usage.used_pages),
        ("SWAP FREE", swap_usage.free_pages),
    ):
        rows.append(
            (
                CellFormat(label, ">"),
                pages,
                binary_units_cell(pages),
                CellFormat(
                    f"{pages / swap_usage.pages if swap_usage.pages else 0:.0%} of TOTAL SWAP",
                    ">",
                ),
            )
        )

    rows.append(())

    commit_limit = vm_commit_limit(prog)
    committed = vm_memory_committed(prog)
    rows.append(
        (
            CellFormat("COMMIT LIMIT", ">"),
            commit_limit,
            binary_units_cell(commit_limit),
            CellFormat("----", "^"),
        )
    )
    rows.append(
        (
            CellFormat("COMMITTED", ">"),
            committed,
            binary_units_cell(committed),
            CellFormat(
                f"{committed / commit_limit:.0%} of TOTAL LIMIT",
                ">",
            ),
        )
    )

    print_table(rows)


def _kmem_vmalloc(
    prog: Program, drgn_arg: bool, *, identified: Optional[IdentifiedVmap] = None
) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.mm import for_each_vmap_area


for va in for_each_vmap_area():
    vm = va.vm
    start = va.va_start
    end = va.va_end
    size = end - start
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            CellFormat("VMAP_AREA", "^"),
            CellFormat("VM_STRUCT", "^"),
            CellFormat("ADDRESS RANGE", "^"),
            CellFormat("SIZE", ">"),
        )
    ]
    if identified is None:

        def areas() -> Iterable[Tuple[Object, Object]]:
            for va in for_each_vmap_area(prog):
                yield va, va.vm.read_()

    else:

        def areas() -> Iterable[Tuple[Object, Object]]:
            yield identified.vmap_area, identified.vm_struct

    for va, vm in areas():
        start = va.va_start.value_()
        end = va.va_end.value_()
        rows.append(
            (
                CellFormat(va.value_(), "^x"),
                CellFormat(vm.value_(), "^x"),
                CellFormat(f"{start:x} - {end:x}", "^"),
                end - start,
            )
        )

    print_table(rows)


def _kmem_vmstat(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.vmstat",
            "global_node_page_state",
            "global_numa_event_state",
            "global_vm_event_state",
            "global_zone_page_state",
        )
        code.append(
            """\
for name, item in prog.type("enum zone_stat_item").enumerators:
    if name == "NR_VM_ZONE_STAT_ITEMS":
        continue
    value = global_zone_page_state(item)

for name, item in prog.type("enum node_stat_item").enumerators:
    if name == "NR_VM_NODE_STAT_ITEMS":
        continue
    value = global_node_page_state(item)


# This depends on CONFIG_NUMA and was added in v4.14.
try:
    numa_stat_item_type = prog.type("enum numa_stat_item")
except LookupError:
    pass
else:
    for name, item in numa_stat_item_type.enumerators:
        # This was renamed in v5.14.
        if name == "NR_VM_NUMA_EVENT_ITEMS" or name == "NR_VM_NUMA_STAT_ITEMS":
            continue
        value = global_numa_event_state(item)

# This depends on CONFIG_VM_EVENT_COUNTERS.
if "vm_event_states" in prog:
    for name, item in prog.type("enum vm_event_item").enumerators:
        if name == "NR_VM_EVENT_ITEMS":
            continue
        value = global_vm_event_state(item)
"""
        )
        code.print()
        return

    rows: List[Sequence[Any]] = []
    rows.append(RowOptions(("VM_ZONE_STAT:",), group=1))
    for name, item in prog.type("enum zone_stat_item").enumerators:  # type: ignore[union-attr]
        if name == "NR_VM_ZONE_STAT_ITEMS":
            continue
        rows.append(
            (
                CellFormat("  " + name, ">"),
                CellFormat(global_zone_page_state(prog, item), "<"),
            )
        )

    rows.append(())
    rows.append(RowOptions(("VM_NODE_STAT:",), group=1))
    for name, item in prog.type("enum node_stat_item").enumerators:  # type: ignore[union-attr]
        if name == "NR_VM_NODE_STAT_ITEMS":
            continue
        rows.append(
            (
                CellFormat("  " + name, ">"),
                CellFormat(global_node_page_state(prog, item), "<"),
            )
        )

    try:
        numa_stat_item_type = prog.type("enum numa_stat_item")
    except LookupError:
        pass
    else:
        rows.append(())
        if "vm_numa_event" in prog:
            rows.append(RowOptions(("VM_NUMA_EVENT:",), group=1))
            nr_numa_items = "NR_VM_NUMA_EVENT_ITEMS"
        else:
            rows.append(RowOptions(("VM_NUMA_STAT:",), group=1))
            nr_numa_items = "NR_VM_NUMA_STAT_ITEMS"
        if nr_numa_items is not None:
            for name, item in numa_stat_item_type.enumerators:  # type: ignore[union-attr]
                if name == nr_numa_items:
                    continue
                rows.append(
                    (
                        CellFormat("  " + name, ">"),
                        CellFormat(global_numa_event_state(prog, item), "<"),
                    )
                )

    if "vm_event_states" in prog:
        rows.append(())
        rows.append(RowOptions(("VM_EVENT_STATES:",), group=1))
        for name, item in prog.type("enum vm_event_item").enumerators:  # type: ignore[union-attr]
            if name == "NR_VM_EVENT_ITEMS":
                continue
            rows.append(
                (
                    CellFormat("  " + name, ">"),
                    CellFormat(global_vm_event_state(prog, item), "<"),
                )
            )

    print_table(rows, sep=": ")


def _kmem_nodes(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.linux.device", "dev_name")
        code.add_from_import(
            "drgn.helpers.linux.mm",
            "PFN_PHYS",
            "decode_memory_block_state",
            "for_each_memory_block",
            "memory_block_size_bytes",
            "pfn_to_page",
        )
        code.add_from_import(
            "drgn.helpers.linux.mmzone",
            "NODE_DATA",
            "decode_section_flags",
            "for_each_online_node",
            "for_each_present_section",
            "section_decode_mem_map",
            "section_mem_map_addr",
            "section_nr_to_pfn",
        )

        code.append(
            """\
for node in for_each_online_node():
    pgdat = NODE_DATA(node)
    size = pgdat.node_spanned_pages
    start_pfn = pgdat.node_start_pfn
    mem_map = pfn_to_page(start_pfn)
    start_paddr = PFN_PHYS(start_pfn)

    for zone in pgdat.node_zones:
        zone_name = zone.name
        zone_size = zone.spanned_pages
        if zone_size == 0:
            continue
        zone_start_pfn = zone.zone_start_pfn.read_()
        zone_mem_map = pfn_to_page(zone_start_pfn)
        zone_start_paddr = PFN_PHYS(zone_start_pfn)


if "mem_section" in prog:  # Check for CONFIG_SPARSEMEM.
    for nr, section in for_each_present_section():
        coded_mem_map = section_mem_map_addr(section)
        mem_map = section_decode_mem_map(section, nr)
        state = decode_section_flags(section)
        pfn = section_nr_to_pfn(nr)


if "memory_subsys" in prog:  # Check for CONFIG_MEMORY_HOTPLUG.
    block_size = memory_block_size_bytes()
    for mem in for_each_memory_block():
        name = dev_name(mem.dev.address_of_())
        start_section_no = mem.start_section_nr
        physical_start = PFN_PHYS(section_nr_to_pfn(start_section_no))
        physical_end = physical_start + block_size
        node = getattr(mem, "nid", None)  # Only available since Linux 4.17.
        state = decode_memory_block_state(mem)
"""
        )
        code.print()
        return

    rows: List[Sequence[Any]] = []
    for nid in for_each_online_node(prog):
        if rows:
            rows.append(RowOptions((), group=3))
            rows.append(RowOptions(("-" * 72,), group=3))
            rows.append(RowOptions((), group=3))

        rows.append(
            (
                CellFormat("NODE", "^"),
                CellFormat("SIZE", "^"),
                CellFormat("PGLIST_DATA", "^"),
                # Crash includes BOOTMEM_DATA, which was removed entirely in
                # Linux kernel commit 355c45affca7 ("mm: remove bootmem
                # allocator implementation.") (in v4.20) and removed from most
                # mainstream architectures years before that. We don't bother.
                CellFormat("NODE_ZONES", "^"),
            )
        )
        pgdat = NODE_DATA(prog, nid)
        prefix: Sequence[Any] = (
            CellFormat(nid, "^"),
            CellFormat(pgdat.node_spanned_pages.value_(), "^"),
            CellFormat(pgdat.value_(), "^x"),
        )
        zone_rows = [
            (),
            RowOptions(
                (
                    CellFormat("ZONE", "^"),
                    "NAME",
                    CellFormat("SIZE", ">"),
                    CellFormat("MEM_MAP", "^"),
                    CellFormat("START_PADDR", ">"),
                    CellFormat("START_PFN", ">"),
                ),
                group=2,
            ),
        ]
        for i, zone in enumerate(pgdat.node_zones):
            rows.append((*prefix, CellFormat(zone.address_, "^x")))
            prefix = ("", "", "")

            zone_size = zone.spanned_pages.value_()
            if zone_size == 0:
                zone_mem_map_cell: Any = 0
                zone_start_paddr_cell: Any = 0
                zone_start_pfn_cell: Any = 0
            else:
                zone_start_pfn = zone.zone_start_pfn.read_()
                zone_mem_map_cell = CellFormat(
                    pfn_to_page(zone_start_pfn).value_(), "x"
                )
                zone_start_paddr_cell = CellFormat(
                    PFN_PHYS(zone_start_pfn).value_(), "x"
                )
                zone_start_pfn_cell = zone_start_pfn.value_()
            zone_rows.append(
                RowOptions(
                    (
                        CellFormat(i, "^"),
                        escape_ascii_string(zone.name.string_(), escape_backslash=True),
                        zone_size,
                        zone_mem_map_cell,
                        zone_start_paddr_cell,
                        zone_start_pfn_cell,
                    ),
                    group=2,
                )
            )

        rows.append(
            RowOptions(
                (
                    CellFormat("MEM_MAP", "^"),
                    CellFormat("START_PADDR", "^"),
                    CellFormat("START_PFN", "^"),
                ),
                group=1,
            )
        )
        node_start_pfn = pgdat.node_start_pfn.read_()
        rows.append(
            RowOptions(
                (
                    CellFormat(pfn_to_page(node_start_pfn).value_(), "^x"),
                    CellFormat(PFN_PHYS(node_start_pfn).value_(), "^x"),
                    CellFormat(node_start_pfn.value_(), "^"),
                ),
                group=1,
            )
        )
        rows.extend(zone_rows)

    print_table(rows)
    rows.clear()

    if "mem_section" in prog:
        print()
        print("-" * 72)
        print()

        rows.append(
            (
                CellFormat("NR", ">"),
                CellFormat("SECTION", "^"),
                CellFormat("CODED_MEM_MAP", "^"),
                CellFormat("MEM_MAP", "^"),
                "STATE",
                "PFN",
            )
        )
        for nr, section in for_each_present_section(prog):
            rows.append(
                (
                    nr,
                    CellFormat(section.value_(), "^x"),
                    CellFormat(section_mem_map_addr(section).value_(), "^x"),
                    CellFormat(section_decode_mem_map(section, nr).value_(), "^x"),
                    "".join(
                        [
                            {
                                "SECTION_MARKED_PRESENT": "P",
                                "SECTION_HAS_MEM_MAP": "M",
                                "SECTION_IS_ONLINE": "O",
                                "SECTION_IS_EARLY": "E",
                                "SECTION_TAINT_ZONE_DEVICE": "D",
                            }.get(flag, "")
                            for flag in decode_section_flags(section).split("|")
                        ]
                    ),
                    CellFormat(section_nr_to_pfn(prog, nr).value_(), "<"),
                )
            )
        print_table(rows)
        rows.clear()

    if "memory_subsys" in prog:
        print()
        print("-" * 72)
        print()

        block_size = memory_block_size_bytes(prog)

        # struct memory_block::nid was added in Linux kernel commit
        # d0dc12e86b31 ("mm/memory_hotplug: optimize memory hotplug") (in
        # v4.17).
        if prog.type("struct memory_block").has_member("nid"):
            node_heading: Sequence[Any] = (CellFormat("NODE", "^"),)
        else:
            node_heading = ()

        rows.append(
            (
                # Crash uses the heading MEM_BLOCK, which can easily be confused
                # with struct memblock, an entirely different type.
                CellFormat("MEMORY_BLOCK", "^"),
                "NAME",
                CellFormat("PHYSICAL RANGE", "^"),
                *node_heading,
                "STATE",
                "START_SECTION_NO",
            )
        )
        paddr_width = len(hex(PFN_PHYS(prog["max_pfn"]))) - 2
        for mem in for_each_memory_block(prog):
            start_section_nr = mem.start_section_nr.read_()
            physical_start = PFN_PHYS(section_nr_to_pfn(start_section_nr))
            physical_end = physical_start + block_size - 1

            if node_heading:
                node_cell: Sequence[Any] = (CellFormat(mem.nid.value_(), "^"),)
            else:
                node_cell = ()

            state = decode_memory_block_state(mem)
            if state.startswith("MEM_"):
                state = state[len("MEM_") :]

            rows.append(
                (
                    CellFormat(mem.value_(), "^x"),
                    escape_ascii_string(
                        dev_name(mem.dev.address_of_()), escape_backslash=True
                    ),
                    CellFormat(
                        f"{physical_start.value_():{paddr_width}x}"
                        f" - {physical_end.value_():{paddr_width}x}",
                        "^",
                    ),
                    *node_cell,
                    state,
                    CellFormat(start_section_nr.value_(), "<"),
                ),
            )
        print_table(rows)


def _kmem_zones(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.mmzone",
            "for_each_online_pgdat",
            "high_wmark_pages",
            "low_wmark_pages",
            "min_wmark_pages",
        )
        code.add_from_import("drgn.helpers.linux.vmstat", "zone_page_state")
        code.append(
            """\
for pgdat in for_each_online_pgdat():
    node = pgdat.node_id
    for zone in pgdat.node_zones:
        zone = zone.address_of_()
        name = zone.name
        size = zone.spanned_pages
        if size == 0:
            continue

        present = zone.present_pages
        min_watermark = min_wmark_pages(zone)
        low_watermark = low_wmark_pages(zone)
        high_watermark = high_wmark_pages(zone)

        for stat_name, stat_item in prog.type("enum zone_stat_item").enumerators:
            if stat_name == "NR_VM_ZONE_STAT_ITEMS":
                continue
            stat_value = zone_page_state(zone, stat_item)
"""
        )
        code.print()
        return

    separator = ""
    for pgdat in for_each_online_pgdat(prog):
        nid = pgdat.node_id.value_()
        for i, zone in enumerate(pgdat.node_zones):
            # Separate with one newline after an unpopulated zone, two after
            # populated.
            sys.stdout.write(separator)

            zone = zone.address_of_()
            print(
                f"NODE: {nid}  ZONE: {i}  ADDR: {zone.value_():x}  NAME: {double_quote_ascii_string(zone.name.string_())}"
            )

            size = zone.spanned_pages.value_()
            if not size:
                print("  [unpopulated]")
                separator = "\n"
                continue

            present = zone.present_pages.value_()
            if present < size:
                present_column = f"  PRESENT: {present}"
            else:
                present_column = ""

            minw = min_wmark_pages(zone)
            loww = low_wmark_pages(zone)
            highw = high_wmark_pages(zone)

            print(
                f"  SIZE: {size}{present_column}  MIN/LOW/HIGH: {minw}/{loww}/{highw}\n  VM_STAT:"
            )

            rows: List[Sequence[Any]] = []
            for name, item in prog.type("enum zone_stat_item").enumerators:  # type: ignore[union-attr]
                if name == "NR_VM_ZONE_STAT_ITEMS":
                    continue
                rows.append(
                    (
                        CellFormat("  " + name, ">"),
                        CellFormat(zone_page_state(zone, item), "<"),
                    )
                )
            print_table(rows)
            separator = "\n\n"


def _kmem_per_cpu_offset(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn import NULL
from drgn.helpers.linux.cpumask import for_each_possible_cpu
from drgn.helpers.linux.percpu import per_cpu_ptr


for cpu in for_each_possible_cpu():
    offset = per_cpu_ptr(NULL(prog, "void *"), cpu)
"""
        )
        return

    print("PER-CPU OFFSET VALUES:")
    nullptr = NULL(prog, "void *")
    for cpu in for_each_possible_cpu(prog):
        cpu_field = f"CPU {cpu}"
        print(f"{cpu_field:>7}: {per_cpu_ptr(nullptr, cpu).value_():x}")


def _kmem_hstate(prog: Program, drgn_arg: bool) -> None:
    if drgn_arg:
        sys.stdout.write(
            """\
from drgn.helpers.linux.hugetlb import for_each_hstate, huge_page_size


for hstate in for_each_hstate(prog):
    size = huge_page_size(hstate)
    free = hstate.free_huge_pages
    total = hstate.nr_huge_pages
    name = hstate.name
"""
        )
        return

    rows: List[Sequence[Any]] = [
        (
            CellFormat("HSTATE", "^"),
            CellFormat("SIZE", ">"),
            CellFormat("FREE", ">"),
            CellFormat("TOTAL", ">"),
            "NAME",
        )
    ]
    for hstate in for_each_hstate(prog):
        rows.append(
            (
                CellFormat(hstate.value_(), "^x"),
                CellFormat(number_in_binary_units(huge_page_size(hstate)) + "B", ">"),
                hstate.free_huge_pages.value_(),
                hstate.nr_huge_pages.value_(),
                escape_ascii_string(hstate.name.string_(), escape_backslash=True),
            )
        )

    print_table(rows)


def _page_flags_and_decoded_flags(page: Object) -> str:
    flags = page_flags(page).read_()
    decoded_flags = decode_page_flags_value(flags).replace("|", ",").replace("PG_", "")
    return f"{flags.value_():x} {decoded_flags}"


def _page_flags_member(page: Object) -> str:
    flags = page_flags(page)
    return f"{flags.value_():0{sizeof(flags) * 2}x}"


def _page_list_head_member(member: str, page: Object) -> str:
    node = page[0].subobject_(member)
    next = node.next
    prev = node.prev
    return f"{next.value_():0{sizeof(next) * 2}x},{prev.value_():0{sizeof(prev) * 2}x}"


def _page_callback_head_member(member: str, page: Object) -> str:
    head = page[0].subobject_(member)
    next = head.next
    func = head.func
    return f"{next.value_():0{sizeof(next) * 2}x},{func.value_():0{sizeof(func) * 2}x}"


def _page_decimal_member(member: str, page: Object) -> str:
    return str(page[0].subobject_(member).value_())


def _page_hex_member(member: str, page: Object) -> str:
    object = page[0].subobject_(member)
    return f"{object.value_():0{sizeof(object) * 2}x}"


_T = TypeVar("_T")


def _print_pages_default_members(
    prog: Program,
    iterable: Iterable[_T],
    *,
    get_page: Callable[[_T], Object],
    get_physical: Optional[Callable[[_T], int]] = None,
    get_mapping: Optional[Callable[[_T], int]] = None,
    get_index: Optional[Callable[[_T], int]] = None,
) -> None:
    address_size = prog.address_size()
    PAGE_SHIFT = prog["PAGE_SHIFT"].value_()

    # print_table() requires having all rows available ahead of time. With a
    # row per page, this would take too much time and memory. Instead, we
    # compute the column widths and print each row manually.
    widths = [
        address_size * 2,
        len(hex(prog["max_pfn"].value_() << PAGE_SHIFT)) - 2,
        address_size * 2,
        12,
        3,
        5,
    ]

    first = True
    for item in iterable:
        if first:
            _print_table_row(
                (
                    CellFormat("PAGE", "^"),
                    CellFormat("PHYSICAL", ">"),
                    CellFormat("MAPPING", "^"),
                    CellFormat("INDEX", ">"),
                    CellFormat("CNT", ">"),
                    "FLAGS",
                ),
                widths=widths,
            )
            first = False

        page = get_page(item)
        _print_table_row(
            (
                f"{page.value_():0{address_size * 2}x}",
                CellFormat(
                    (
                        page_to_pfn(page).value_() << PAGE_SHIFT
                        if get_physical is None
                        else get_physical(item)
                    ),
                    "x",
                ),
                CellFormat(
                    page.mapping.value_() if get_mapping is None else get_mapping(item),
                    "x",
                ),
                CellFormat(
                    page_index(page).value_() if get_index is None else get_index(item),
                    "x",
                ),
                CellFormat(page._refcount.counter.value_(), "x"),
                _page_flags_and_decoded_flags(page),
            ),
            widths=widths,
        )


def _print_pages_members(
    prog: Program,
    members: str,
    pages: Iterable[Object],
) -> None:
    address_size = prog.address_size()
    header: List[Any] = [CellFormat("PAGE", "^")]
    widths = [address_size * 2]
    getters: List[Callable[[Object], Any]] = [
        lambda page: f"{page.value_():0{address_size * 2}x}"
    ]

    struct_page = prog.type("struct page")
    integer_base = prog.config.get("crash_radix", 10)
    for member in _parse_members(members):
        header.append(member)

        if member == "flags":
            widths.append(address_size * 2)
            getters.append(_page_flags_member)
            continue

        member_type = typeof_member(struct_page, member)
        member_type_kind = member_type.unaliased_kind()
        member_type_name = member_type.type_name()
        if member_type_name == "struct list_head":
            widths.append(max(address_size * 4 + 1, len(member)))
            getters.append(functools.partial(_page_list_head_member, member))
        elif member_type_name == "struct callback_head":
            widths.append(max(address_size * 4 + 1, len(member)))
            getters.append(functools.partial(_page_callback_head_member, member))
        elif member_type_name == "atomic_t" or member_type_name == "atomic_long_t":
            widths.append(max(12, len(member)))
            getters.append(functools.partial(_page_decimal_member, member + ".counter"))
        elif (
            member_type_kind == TypeKind.INT and integer_base == 16
        ) or member_type_kind == TypeKind.POINTER:
            widths.append(max(sizeof(member_type) * 2, len(member)))
            getters.append(functools.partial(_page_hex_member, member))
        elif member_type_kind == TypeKind.INT:
            widths.append(max(12, len(member)))
            getters.append(functools.partial(_page_decimal_member, member))
        else:
            raise NotImplementedError(
                f"formatting {member_type_name!r} not implemented"
            )

    _print_table_row(header, widths=widths)
    for page in pages:
        _print_table_row([getter(page) for getter in getters], widths=widths)


def _kmem_pages(
    prog: Program,
    drgn_arg: bool,
    *,
    members: Optional[str] = None,
    identified: Optional[IdentifiedPage] = None,
) -> None:
    if members is None:
        if drgn_arg:
            code = CrashDrgnCodeBuilder(prog)
            code.add_from_import(
                "drgn.helpers.linux.mm",
                "PFN_PHYS",
                "decode_page_flags_value",
                "for_each_valid_pfn_and_page",
                "page_flags",
                "page_index",
            )
            code.append(
                """\
for pfn, page in for_each_valid_pfn_and_page():
    physical = PFN_PHYS(pfn)
    mapping = page.mapping
    index = page_index(page)
    cnt = page._refcount.counter
    flags = page_flags(page)
    decoded_flags = decode_page_flags_value(flags)
"""
            )
            return code.print()

        PAGE_SHIFT = prog["PAGE_SHIFT"].value_()
        _print_pages_default_members(
            prog,
            (
                for_each_valid_pfn_and_page(prog)
                if identified is None
                else ((identified.pfn, identified.page),)
            ),
            get_page=operator.itemgetter(1),
            get_physical=lambda item: item[0] << PAGE_SHIFT,
        )
    else:
        if drgn_arg:
            code = CrashDrgnCodeBuilder(prog)
            code.add_from_import("drgn.helpers.linux.mm", "for_each_page")
            code.append("for page in for_each_page():\n")
            for member in _parse_members(members):
                if member == "flags":
                    code.add_from_import("drgn.helpers.linux.mm", "page_flags")
                    code.append("    flags = page_flags(page)\n")
                else:
                    code.append(
                        f"    {_sanitize_member_name(member)} = " f"page.{member}\n"
                    )
            return code.print()

        _print_pages_members(
            prog,
            members,
            for_each_page(prog) if identified is None else (identified.page,),
        )


def _kmem_slab(
    prog: Program,
    drgn_arg: bool,
    *,
    ignore: AbstractSet[bytes],
    names: Optional[Sequence[str]] = None,
    identified: Optional[IdentifiedSlabObject] = None,
) -> None:
    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import(
            "drgn.helpers.linux.slab", "slab_cache_pages_per_slab", "slab_cache_usage"
        )

        if len(ignore) > 1:
            code.append("ignore = {")
            code.append(", ".join([repr(name) for name in sorted(ignore)]))
            code.append("}\n\n")

        if names:
            code.add_from_import("drgn.helpers.linux.slab", "find_slab_cache")
            # If we're only matching one name, use an if statement instead of a
            # loop. However, ignore is handled with a continue statement, so
            # always use a loop if there are names to ignore.
            if len(names) == 1 and not ignore:
                code.append(
                    f"""\
cache = find_slab_cache({_repr_black(names[0])})
if cache:
"""
                )
            else:
                code.append("for search_name in (\n")
                for search_name in names:
                    code.append(f"    {_repr_black(search_name)},\n")
                code.append(
                    """):
    cache = find_slab_cache(search_name)
    if not cache:
        continue
"""
                )
        else:
            code.add_from_import("drgn.helpers.linux.slab", "for_each_slab_cache")
            code.append("for cache in for_each_slab_cache():\n")

        code.append("    name = cache.name\n")

        if ignore:
            if len(ignore) == 1:
                code.append(f"    if name.string_() == {next(iter(ignore))!r}:\n")
            else:
                code.append("    if name.string_() in ignore:\n")
            code.append("        continue\n")

        code.append(
            """\
    objsize = cache.object_size
    try:
        usage = slab_cache_usage(cache)
    except ValueError:
        # SLUB without SLUB_DEBUG and SLOB do not support slab_cache_usage().
        pass
    else:
        allocated = usage.active_objs
        total = usage.num_objs
        slabs = usage.num_slabs
    try:
        ssize = prog["PAGE_SIZE"] * slab_cache_pages_per_slab(cache)
    except ValueError:
        # SLOB does not support slab_cache_pages_per_slab().
        pass
"""
        )
        code.print()
        return

    rows: List[Sequence[Any]] = [
        (
            "CACHE",
            CellFormat("OBJSIZE", ">"),
            CellFormat("ALLOCATED", ">"),
            CellFormat("TOTAL", ">"),
            CellFormat("SLABS", ">"),
            CellFormat("SSIZE", ">"),
            "NAME",
        )
    ]

    if identified is not None:
        caches: Iterable[Object] = (identified.slab_object_info.slab_cache,)
    elif names is None:
        caches = for_each_slab_cache(prog)
    else:
        caches = [cache for name in names if (cache := find_slab_cache(prog, name))]

    for cache in caches:
        name = cache.name.string_()
        if name in ignore:
            rows.append(
                (
                    CellFormat(cache.value_(), "<x"),
                    "[IGNORED]",
                    "",
                    "",
                    "",
                    "",
                    escape_ascii_string(cache.name.string_(), escape_backslash=True),
                )
            )
        else:
            objsize = cache.object_size.value_()
            try:
                order = slab_cache_order(cache)
            except ValueError:
                # SLOB doesn't support slab_cache_order() or
                # slab_cache_usage().
                allocated: Any = "[UNKNOWN]"
                total: Any = ""
                slabs: Any = ""
                ssize: Any = ""
            else:
                ssize = CellFormat(
                    f"{(prog['PAGE_SIZE'].value_() << order) // 1024}k", ">"
                )
                # SLUB without SLUB_DEBUG supports slab_cache_order() but not
                # slab_cache_usage().
                try:
                    usage = slab_cache_usage(cache)
                except ValueError:
                    allocated = "[UNKNOWN]"
                    total = ""
                    slabs = ""
                except (FaultError, ValidationError):
                    allocated = "[CORRUPTED]"
                    total = ""
                    slabs = ""
                else:
                    allocated = usage.active_objs
                    total = usage.num_objs
                    slabs = usage.num_slabs
            rows.append(
                (
                    CellFormat(cache.value_(), "<x"),
                    objsize,
                    allocated,
                    total,
                    slabs,
                    ssize,
                    escape_ascii_string(cache.name.string_(), escape_backslash=True),
                )
            )

            if identified is not None:
                rows.append(
                    RowOptions(
                        (
                            "",
                            "SLAB",
                            "MEMORY",
                        ),
                        group=1,
                    )
                )
                slab = identified.slab_object_info.slab
                slab_memory = page_to_virt(cast("struct page *", slab))
                rows.append(
                    RowOptions(
                        (
                            "",
                            f"{slab.value_():x}",
                            f"{slab_memory.value_():x}",
                        ),
                        group=1,
                    )
                )

                rows.append(
                    RowOptions(
                        (
                            "",
                            "FREE / [ALLOCATED]",
                        ),
                        group=2,
                    )
                )
                object_address = f"{identified.slab_object_info.address:x}"
                if identified.slab_object_info.allocated:
                    object_address = f"[{object_address}]"
                rows.append(
                    RowOptions(
                        (
                            "",
                            object_address,
                        ),
                        group=2,
                    )
                )
    print_table(rows)


def _kmem_page_flags(prog: Program, drgn_arg: bool, flags: Optional[int]) -> None:
    if drgn_arg:
        if flags is None:
            sys.stdout.write(
                """\
for name, bit in prog.type("enum pageflags").enumerators:
    if name == "__NR_PAGEFLAGS":
        continue
    value = 1 << bit
"""
            )
        else:
            sys.stdout.write(
                f"""\
from drgn.helpers.linux.mm import decode_page_flags_value


flags = decode_page_flags_value(0x{flags:x})
"""
            )
        return

    if flags is None:
        prefix = ""
    else:
        print(f"FLAGS: {flags:x}")
        prefix = "  "
    rows: List[Sequence[Any]] = [
        (prefix + "PAGE-FLAG", CellFormat("BIT", ">"), "VALUE")
    ]
    width = len(hex((1 << prog["__NR_PAGEFLAGS"].value_()) - 1)) - 2
    for name, bit in prog.type("enum pageflags").enumerators:  # type: ignore[union-attr]
        if name == "__NR_PAGEFLAGS":
            continue
        value = 1 << bit
        if flags is None or flags & value:
            rows.append((prefix + name, bit, f"{value:0{width}x}"))
    print_table(rows)


_IDENTIFY_MODES = {
    None,
    # Crash supposedly supports -f, too, but I've never been able to get it to
    # work, so we ignore it for now.
    "free",
    "pages",
    "slab",
    "vmalloc",
}


def _kmem_identify(
    prog: Program,
    drgn_arg: bool,
    addresses: List[int],
    mode: Optional[str],
    *,
    page_members: Optional[str],
    ignore_slab_caches: AbstractSet[bytes],
    slab_cache_names: bool,
) -> None:
    if mode not in _IDENTIFY_MODES:
        raise CommandArgumentError(f"address not allowed with {mode}")

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn.helpers.common.memory", "identify_address")
        if len(addresses) == 1:
            code.append(f"address = {hex(addresses[0])}\n")
        else:
            code.append("for address in (")
            code.append(", ".join([hex(address) for address in addresses]))
            code.append("):\n    ")
        code.append("identified = identify_address(address)\n")
        code.print()
        return

    first = True

    def print_divider() -> None:
        nonlocal first
        if first:
            first = False
        else:
            print()

    for address in addresses:
        all_identified = identify_address_all(prog, address)
        if all_identified is None:
            all_identified = ()

        found_vmap = False
        found_page = False
        found_slab = False
        for identified in all_identified:
            if isinstance(identified, IdentifiedSymbol):
                if mode is None:
                    offset = address - identified.symbol.address
                    offset_str = f"+{offset}" if offset else ""

                    module_str = ""
                    try:
                        module = prog.module(address)
                    except LookupError:
                        pass
                    else:
                        if isinstance(module, RelocatableModule):
                            module_str = f" [{module.name}]"

                    print_divider()
                    print(
                        f"{address:x} (?) {identified.symbol.name}{offset_str}{module_str}"
                    )
            elif isinstance(identified, (IdentifiedTaskStruct, IdentifiedTaskStack)):
                if mode is None:
                    print_divider()
                    _print_sys(prog, system_fields=False, context=identified.task)
            elif isinstance(identified, IdentifiedVmap):
                found_vmap = True
                if mode is None or mode == "vmalloc":
                    print_divider()
                    _kmem_vmalloc(prog, False, identified=identified)
            elif isinstance(identified, IdentifiedPage):
                found_page = True
                if mode is None or mode == "pages":
                    print_divider()
                    _kmem_pages(
                        prog, False, members=page_members, identified=identified
                    )
            elif isinstance(identified, IdentifiedSlabObject):
                found_slab = True
                if mode is None or mode == "slab":
                    print_divider()
                    if slab_cache_names:
                        print(
                            f"kmem: ignoring pre-selected slab caches for address: {address:x}"
                        )
                    if identified.slab_object_info.address:
                        _kmem_slab(
                            prog,
                            False,
                            ignore=ignore_slab_caches,
                            identified=identified,
                        )
                    else:
                        print(f"kmem: address is from SLOB: {address:x}")

        if not found_vmap and mode == "vmalloc":
            print_divider()
            # Crash fails with a bad memory access instead.
            print(f"kmem: address is not allocated in vmalloc subsystem: {address:x}")

        if not found_page and (mode is None or mode == "pages"):
            try:
                pfn = follow_pfn(prog["init_mm"].address_of_(), address)
            except (FaultError, NotImplementedError):
                pass
            else:
                print_divider()
                _kmem_pages(
                    prog,
                    False,
                    members=page_members,
                    identified=IdentifiedPage(address, pfn_to_page(pfn), pfn.value_()),
                )

        if mode == "slab" and not found_slab:
            print_divider()
            print(f"kmem: address is not allocated in slab subsystem: {address:x}")


@crash_command(
    description="kernel memory",
    long_description="""
    Display information about various parts of the memory management subsystem.
    """,
    arguments=(
        mutually_exclusive_group(
            argument(
                "-f",
                dest="mode",
                action="store_const",
                const="free",
                help="display and verify page allocator free lists",
            ),
            argument(
                "-F",
                dest="mode",
                action="store_const",
                const="free_pages",
                help="like -f, but also display each page on the free lists",
            ),
            argument(
                "-i",
                dest="mode",
                action="store_const",
                const="info",
                help="display general memory usage information",
            ),
            argument(
                "-v",
                dest="mode",
                action="store_const",
                const="vmalloc",
                help="display memory regions allocated with vmalloc()/vmap()",
            ),
            argument(
                "-V",
                dest="mode",
                action="store_const",
                const="vmstat",
                help="display zone, node, NUMA, and VM event statistics",
            ),
            argument(
                "-n",
                dest="mode",
                action="store_const",
                const="nodes",
                help="display NUMA nodes, SPARSEMEM sections, and memory blocks",
            ),
            argument(
                "-z",
                dest="mode",
                action="store_const",
                const="zones",
                help="display per-zone memory statistics",
            ),
            argument(
                "-o",
                dest="mode",
                action="store_const",
                const="per_cpu_offset",
                help="""
                display each CPU's per-CPU offset (the value added to convert a
                per-CPU symbol to a virtual address)
                """,
            ),
            argument(
                "-h",
                dest="mode",
                action="store_const",
                const="hstate",
                help="display HugeTLB state",
            ),
            argument(
                "-p",
                dest="mode",
                action="store_const",
                const="pages",
                help="""
                display every valid struct page, including its physical
                address, mapping, index, refcount, and flags
                """,
            ),
            argument(
                "-m",
                dest="page_members",
                help="""
                display the given comma-separated members of every valid struct page
                """,
            ),
            argument(
                "-s",
                dest="mode",
                action="store_const",
                const="slab",
                help="display an overview of slab caches",
            ),
            argument(
                "-g",
                dest="page_flags",
                metavar="FLAGS",
                type="hexadecimal",
                nargs="?",
                default=argparse.SUPPRESS,
                help="""
                display the flags set on a hexadecimal page flags value, or
                display all of the possible flags if no value is given
                """,
            ),
        ),
        argument(
            "-I",
            dest="ignore_slab_caches",
            metavar="NAME[,NAME...]",
            help="""
            when used with -s, comma-separated list of names of slab caches to
            ignore
            """,
        ),
        argument(
            "name",
            nargs="*",
            help="""
            when used with -s, only display the slab caches with the given names
            """,
        ),
        argument(
            "address",
            nargs="*",
            help="""
            addresses to identify as symbols, tasks, task stacks, vmalloc
            allocations, pages, and/or slab objects. Can be combined with -v,
            -p, -m, or -s to limit the search to their respective types
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_kmem(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    # argparse greedily assigns all of the positional arguments to name.
    # Separate the addresses.
    posargs = args.name
    args.name = []
    assert not args.address
    for posarg in posargs:
        try:
            args.address.append(int(posarg, 16))
        except ValueError:
            args.name.append(posarg)

    if args.page_members is not None:
        args.mode = "pages"
    elif hasattr(args, "page_flags"):
        args.mode = "page_flags"

    if args.ignore_slab_caches is None:
        ignore_slab_caches: AbstractSet[bytes] = frozenset()
    else:
        if args.mode != "slab":
            raise CommandArgumentError("-I can only be used with -s")
        ignore_slab_caches = {
            name.encode() for name in args.ignore_slab_caches.split(",")
        }

    if args.name:
        if args.mode != "slab":
            raise CommandArgumentError("name can only be used with -s")
        slab_cache_names: Optional[List[str]] = args.name
    else:
        slab_cache_names = None

    if args.address:
        _kmem_identify(
            prog,
            args.drgn,
            args.address,
            args.mode,
            page_members=args.page_members,
            ignore_slab_caches=ignore_slab_caches,
            slab_cache_names=bool(slab_cache_names),
        )
    elif args.mode == "free":
        return _kmem_free(prog, args.drgn)
    elif args.mode == "free_pages":
        return _kmem_free(prog, args.drgn, show_pages=True)
    elif args.mode == "info":
        return _kmem_info(prog, args.drgn)
    elif args.mode == "vmalloc":
        return _kmem_vmalloc(prog, args.drgn)
    elif args.mode == "vmstat":
        return _kmem_vmstat(prog, args.drgn)
    elif args.mode == "nodes":
        return _kmem_nodes(prog, args.drgn)
    elif args.mode == "zones":
        return _kmem_zones(prog, args.drgn)
    elif args.mode == "per_cpu_offset":
        return _kmem_per_cpu_offset(prog, args.drgn)
    elif args.mode == "hstate":
        return _kmem_hstate(prog, args.drgn)
    elif args.mode == "pages":
        return _kmem_pages(prog, args.drgn, members=args.page_members)
    elif args.mode == "slab":
        return _kmem_slab(
            prog, args.drgn, ignore=ignore_slab_caches, names=slab_cache_names
        )
    elif args.mode == "page_flags":
        return _kmem_page_flags(prog, args.drgn, args.page_flags)
    else:
        raise CommandArgumentError("address or option is required")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Signals
-------

The ``drgn.helpers.linux.signal`` module provides helpers for working with
signals.
"""

import operator
import types
from typing import Dict, Iterator, List, Mapping, Sequence, Union

import _drgn_util.platform
from drgn import IntegerLike, Object, Program, TypeKind, sizeof
from drgn.helpers.common.format import decode_flags
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.bitops import for_each_set_bit
from drgn.helpers.linux.list import list_for_each_entry

__all__ = (
    "decode_sigaction_flags",
    "decode_sigaction_flags_value",
    "decode_sigset",
    "sigaction_flags",
    "signal_names",
    "signal_numbers",
    "sigpending_for_each",
    "sigset_to_hex",
)


def sigpending_for_each(pending: Object) -> Iterator[Object]:
    """
    Iterate over a queue of pending signals.

    :param pending: ``struct sigpending *``
    :return: Iterator of ``struct sigqueue *`` objects.
    """
    return list_for_each_entry("struct sigqueue", pending.list.address_of_(), "list")


@takes_program_or_default
def signal_names(prog: Program) -> Mapping[str, int]:
    """
    Get a mapping from signal names to numbers on this kernel.

    >>> signal_names()
    {'SIGHUP': 1, 'SIGINT': 2, 'SIGQUIT': 3, ...}

    If there are multiple names for the same number (e.g., ``SIGCHLD`` and
    ``SIGCLD``), then the preferred name comes first.

    Note that ``SIGRTMIN`` and ``SIGRTMAX`` are dynamic in userspace, so the
    values used in the kernel may differ from those in userspace.
    """
    return _drgn_util.platform.SIGNALS_BY_MACHINE_NAME[
        prog.platform.arch.name.lower()  # type: ignore[union-attr]
    ]


@takes_program_or_default
def signal_numbers(prog: Program) -> Mapping[int, Sequence[str]]:
    """
    Get a mapping from signal numbers to names on this kernel.

    >>> signal_numbers()
    {1: ['SIGHUP'], 2: ['SIGINT'], 3: ['SIGQUIT'], ...}

    If there are multiple names for the same number (e.g., ``SIGCHLD`` and
    ``SIGCLD``), then the preferred name comes first.

    Note that ``SIGRTMIN`` and ``SIGRTMAX`` are dynamic in userspace, so the
    values used in the kernel may differ from those in userspace.
    """
    try:
        return prog.cache["signal_numbers"]
    except KeyError:
        pass

    name_to_number = signal_names(prog)
    number_to_names: Dict[int, List[str]] = {}
    for name, number in name_to_number.items():
        if name == "SIGRTMAX":
            rtmin = name_to_number["SIGRTMIN"]
            # This is how crash numbers real-time signals, maybe inherited from
            # the kill -l shell builtin on ksh:
            # https://git.kernel.org/pub/scm/utils/dash/dash.git/tree/src/mksignames.c?h=v0.5.13#n80.
            for rtnumber in range(rtmin + 1, (rtmin + number) // 2 + 1):
                number_to_names.setdefault(rtnumber, []).append(
                    f"SIGRTMIN+{rtnumber - rtmin}"
                )
            for rtnumber in range((rtmin + number) // 2 + 1, number):
                number_to_names.setdefault(rtnumber, []).append(
                    f"SIGRTMAX-{number - rtnumber}"
                )
        number_to_names.setdefault(number, []).append(name)
    result = types.MappingProxyType(number_to_names)
    prog.cache["signal_numbers"] = result
    return result


@takes_program_or_default
def sigaction_flags(prog: Program) -> Mapping[str, int]:
    """
    Get a mapping of :manpage:`sigaction(2)` ``sa_flags`` flag names to values
    on this kernel.

    >>> sigaction_flags()
    {'SA_NOCLDSTOP': 1, 'SA_NOCLDWAIT': 2, 'SA_SIGINFO': 4, ...}

    If there are multiple names for the same value (e.g., ``SA_NODEFER`` and
    ``SA_NOMASK``), then the preferred name comes first.
    """
    return _drgn_util.platform.SIGACTION_FLAGS_BY_MACHINE_NAME[
        prog.platform.arch.name.lower()  # type: ignore[union-attr]
    ]


@takes_program_or_default
def decode_sigset(prog: Program, sigset: Union[Object, IntegerLike]) -> str:
    """
    Get a human-readable representation of a signal set.

    >>> decode_sigset(task.blocked)
    '{SIGINT,SIGTERM}'
    >>> decode_sigset(0x14003)
    '{SIGHUP,SIGINT,SIGTERM,SIGCHLD}'

    :param sigset: ``sigset_t``, ``sigset_t *``, or an integer
    """
    if not isinstance(sigset, Object) or sigset.type_.unaliased_kind() == TypeKind.INT:
        sigset_int = operator.index(sigset)
        if sigset_int < 0:
            raise ValueError("sigset cannot be negative")
        sig_words = []
        sigset_type = prog.type("sigset_t")
        NSIG_BPW: int = sigset_type.member("sig").type.type.size * 8  # type: ignore[operator]
        mask = (1 << NSIG_BPW) - 1
        while sigset_int:
            sig_words.append(sigset_int & mask)
            sigset_int >>= NSIG_BPW
        sigset = Object(prog, sigset_type, {"sig": sig_words})

    number_to_names = signal_numbers(sigset.prog_)
    sig = sigset.sig
    names = [
        number_to_names[bit + 1][0] for bit in for_each_set_bit(sig, sizeof(sig) * 8)
    ]
    return f"{{{','.join(names)}}}"


def sigset_to_hex(sigset: Object) -> str:
    """
    Get a hexadecimal representation of a signal set.

    This matches how signal sets are represented in
    :manpage:`proc_pid_status(5)`.

    >>> sigset_to_hex(task.blocked)
    '0000000000004002'

    :param sigset: ``sigset_t`` or ``sigset_t *``
    """
    NSIG: int = sigset.prog_.type("struct sighand_struct").member("action").type.length  # type: ignore[assignment]
    sig = sigset.sig
    NSIG_BPW: int = sig.type_.type.size * 8  # type: ignore[operator]
    max_width = NSIG_BPW // 4
    if NSIG % NSIG_BPW == 0:
        width = max_width
    else:
        width = (NSIG % NSIG_BPW + 3) // 4

    parts = []
    for word in reversed(sig.value_()):
        parts.append(f"{word:0{width or ''}x}")
        width = max_width
    return "".join(parts)


def decode_sigaction_flags(sigaction: Object) -> str:
    """
    Get a human-readable representation of the ``sa_flags`` of a
    ``struct sigaction``.

    >>> decode_sigaction_flags(task.sighand.action[1].sa)
    'SA_RESTORER|SA_RESTART'

    :param sigaction: ``struct sigaction``
    """
    return decode_sigaction_flags_value(sigaction.prog_, sigaction.sa_flags.value_())


@takes_program_or_default
def decode_sigaction_flags_value(prog: Program, value: IntegerLike) -> str:
    """
    Get a human-readable representation of a :manpage:`sigaction(2)`
    ``sa_flags`` value.

    >>> decode_sigaction_flags_value(0x14000000)
    'SA_RESTORER|SA_RESTART'

    :param value: ``int``
    """
    return decode_flags(
        value, sigaction_flags(prog).items(), bit_numbers=False, aliases=False
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Helpers
-------

The ``drgn.helpers`` package contains subpackages which provide helpers for
working with particular types of programs. Currently, there are common helpers
and helpers for the Linux kernel. In the future, there may be helpers for,
e.g., glibc and libstdc++.
"""


class ValidationError(Exception):
    """
    Error raised by a validator when an inconsistent or invalid state is
    detected.

    Validators are a special category of helpers that check the consistency of
    a data structure. In general, helpers assume that the data structures that
    they examine are valid. Validators do not make this assumption and do
    additional (potentially expensive) checks to detect broken invariants,
    corruption, etc.

    Validators raise :class:`drgn.helpers.ValidationError` if the data
    structure is not valid or :class:`drgn.FaultError` if the data structure is
    invalid in a way that causes a bad memory access. They have names prefixed
    with ``validate_``.
    """



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Memory Zones
------------

The ``drgn.helpers.linux.mmzone`` module provides helpers for working with
memory zones and `SPARSEMEM <https://docs.kernel.org/mm/memory-model.html>`_.
"""

import operator
from typing import Iterator, Mapping, Tuple

from drgn import NULL, IntegerLike, Object, ObjectNotFoundError, Program, TypeKind
from drgn.helpers.common.format import decode_flags
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.nodemask import for_each_online_node

__all__ = (
    "NODE_DATA",
    "decode_section_flags",
    "early_section",
    "early_section_nr",
    "for_each_online_pgdat",
    "for_each_present_section",
    "high_wmark_pages",
    "low_wmark_pages",
    "min_wmark_pages",
    "nr_to_section",
    "online_section",
    "online_section_nr",
    "pfn_to_section",
    "pfn_to_section_nr",
    "present_section",
    "present_section_nr",
    "section_decode_mem_map",
    "section_mem_map_addr",
    "section_nr_to_pfn",
    "valid_section",
    "valid_section_nr",
    "wmark_pages",
)


@takes_program_or_default
def NODE_DATA(prog: Program, nid: IntegerLike) -> Object:
    """
    Get the NUMA node memory layout data of a given NUMA node.

    :param nid: NUMA node ID.
    :return: ``struct pglist_data *``
    """
    try:
        return prog["node_data"][nid]
    except ObjectNotFoundError:
        # CONFIG_NUMA=n
        return prog["contig_page_data"].address_of_()


@takes_program_or_default
def for_each_online_pgdat(prog: Program) -> Iterator[Object]:
    """
    Get the NUMA node memory layout data of each online NUMA node.

    :return: Iterator of ``struct pglist_data *`` objects
    """
    for nid in for_each_online_node(prog):
        yield NODE_DATA(prog, nid)


def min_wmark_pages(zone: Object) -> int:
    """
    Return the given memory zone's minimum watermark.

    This controls when *direct reclaim* is performed.

    :param zone: ``struct zone *``
    """
    return wmark_pages(zone, zone.prog_["WMARK_MIN"])


def low_wmark_pages(zone: Object) -> int:
    """
    Return the given memory zone's low watermark.

    This controls when *indirect reclaim* is performed.

    :param zone: ``struct zone *``
    """
    return wmark_pages(zone, zone.prog_["WMARK_LOW"])


def high_wmark_pages(zone: Object) -> int:
    """
    Return the given memory zone's high watermark.

    This controls when ``kswapd`` can go to sleep.

    :param zone: ``struct zone *``
    """
    return wmark_pages(zone, zone.prog_["WMARK_HIGH"])


def wmark_pages(zone: Object, i: IntegerLike) -> int:
    """
    Return the given watermark in a memory zone.

    >>> wmark_pages(zone, prog["WMARK_PROMO"])
    28

    :param zone: ``struct zone *``
    :param i: ``enum zone_watermarks``
    """
    # Linux kernel commit a921444382b4 ("mm: move zone watermark accesses
    # behind an accessor") (in v5.0) renamed the member, and then Linux kernel
    # commit 1c30844d2dfe ("mm: reclaim small amounts of memory when an
    # external fragmentation event occurs") (in v5.0) added struct
    # zone::watermark_boost.
    try:
        wmark = zone._watermark[i]
    except AttributeError:
        wmark = zone.watermark[i]
    else:
        wmark += zone.watermark_boost
    return wmark.value_()


@takes_program_or_default
def nr_to_section(prog: Program, nr: IntegerLike) -> Object:
    """
    Get the SPARSEMEM section with the given number.

    :param nr: ``unsigned long``
    :return: ``struct mem_section *`` (``NULL`` if not found)
    """
    nr = operator.index(nr)
    SECTIONS_PER_ROOT = prog["SECTIONS_PER_ROOT"].value_()
    root = nr // SECTIONS_PER_ROOT

    if root >= prog["NR_SECTION_ROOTS"].value_():
        return NULL(prog, "struct mem_section *")

    mem_section = prog["mem_section"]
    unaliased_type = mem_section.type_.unaliased()
    if unaliased_type.kind == TypeKind.POINTER:
        mem_section = mem_section.read_()
        if not mem_section:
            return NULL(prog, unaliased_type.type)

    mem_section_root = mem_section[root]
    if mem_section_root.type_.unaliased_kind() == TypeKind.POINTER:
        mem_section_root = mem_section_root.read_()
        if not mem_section_root:
            return mem_section_root

    return mem_section_root + (nr & (SECTIONS_PER_ROOT - 1))


@takes_program_or_default
def pfn_to_section_nr(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the SPARSEMEM section number containing the given page frame number
    (PFN).

    :param pfn: ``unsigned long``
    :return: ``unsigned long``
    """
    return Object(prog, "unsigned long", pfn) >> (
        prog["SECTION_SIZE_BITS"] - prog["PAGE_SHIFT"]
    )


@takes_program_or_default
def section_nr_to_pfn(prog: Program, nr: IntegerLike) -> Object:
    """
    Get the first page frame number (PFN) in the given SPARSEMEM section
    number.

    :param nr: ``unsigned long``
    :return: ``unsigned long``
    """
    return Object(prog, "unsigned long", nr) << (
        prog["SECTION_SIZE_BITS"] - prog["PAGE_SHIFT"]
    )


@takes_program_or_default
def pfn_to_section(prog: Program, pfn: IntegerLike) -> Object:
    """
    Get the SPARSEMEM section containing the given page frame number (PFN).

    :param pfn: ``unsigned long``
    :return: ``struct mem_section *``
    """
    return nr_to_section(pfn_to_section_nr(prog, pfn))


def _section_flags(prog: Program) -> Mapping[str, int]:
    try:
        return prog.cache["section_flags"]
    except KeyError:
        pass

    # Since Linux kernel commit ed7802dd48f7 ("mm: memory_hotplug: enumerate
    # all supported section flags") (in v6.0), the section flags are an
    # anonymous enum. Before that, they were macros.
    try:
        flags_type = prog["SECTION_MARKED_PRESENT_BIT"].type_
    except ObjectNotFoundError:
        flags = {
            # These haven't changed since they were introduced in Linux kernel
            # commit 29751f6991e8 ("[PATCH] sparsemem hotplug base") (in
            # v2.6.13).
            "SECTION_MARKED_PRESENT": 0x1,
            "SECTION_HAS_MEM_MAP": 0x2,
            # Strictly speaking, the remaining flags didn't exit in older
            # kernel versions, but the bits won't be set before they were
            # introduced anyways.
            # Introduced in Linux kernel commit 2d070eab2e82 ("mm: consider
            # zone which is not fully populated to have holes") (in v4.13).
            "SECTION_IS_ONLINE": 0x4,
            # Introduced in Linux kernel commit 326e1b8f83a4 ("mm/sparsemem:
            # introduce a SECTION_IS_EARLY flag") (in v5.3).
            "SECTION_IS_EARLY": 0x8,
            # Introduced in Linux kernel commit 1f90a3477df3 ("mm: teach
            # pfn_to_online_page() about ZONE_DEVICE section collisions") (in
            # v5.12).
            "SECTION_TAINT_ZONE_DEVICE": 0x10,
            "SECTION_MAP_MASK": 0xFFFFFFFFFFFFFFE0,
        }
    else:
        flags = {}
        for enumerator in flags_type.enumerators:  # type: ignore[union-attr]
            name = enumerator.name
            if name == "SECTION_MAP_LAST_BIT":
                flags["SECTION_MAP_MASK"] = (
                    (1 << enumerator.value) - 1
                ) ^ 0xFFFFFFFFFFFFFFFF
            elif name.endswith("_BIT"):
                flags[name[: -len("_BIT")]] = 1 << enumerator.value

    prog.cache["section_flags"] = flags
    return flags


def section_mem_map_addr(section: Object) -> Object:
    """
    Get the SPARSEMEM memory map for the given section.

    This is encoded such that ``mem_map[pfn]`` is the ``struct page`` for the
    given page frame number (PFN).

    :param section: ``struct mem_section *``
    :return: ``struct page *``
    """
    prog = section.prog_
    return Object(
        prog,
        "struct page *",
        section.section_mem_map.value_() & _section_flags(prog)["SECTION_MAP_MASK"],
    )


def section_decode_mem_map(section: Object, nr: IntegerLike) -> Object:
    """
    Get the decoded address of the SPARSEMEM memory map for the given section.

    This is the address such that ``mem_map[0]`` is the ``struct page`` for the
    first page in the section.

    :param section: ``struct mem_section *``
    :param nr: Section number.
    :return: ``struct page *``
    """
    return section_mem_map_addr(section) + section_nr_to_pfn(section.prog_, nr)


def present_section(section: Object) -> bool:
    """
    Return whether a SPARSEMEM section is present.

    :param section: ``struct mem_section *``
    """
    section = section.read_()
    if not section:
        return False
    return bool(
        section.section_mem_map.value_()
        & _section_flags(section.prog_)["SECTION_MARKED_PRESENT"]
    )


@takes_program_or_default
def present_section_nr(prog: Program, nr: IntegerLike) -> bool:
    """
    Return whether the SPARSEMEM section with the given number is present.

    :param nr: ``unsigned long``
    """
    return present_section(nr_to_section(prog, nr))


def valid_section(section: Object) -> bool:
    """
    Return whether a SPARSEMEM section is valid, i.e., has a ``mem_map``.

    :param section: ``struct mem_section *``
    """
    section = section.read_()
    if not section:
        return False
    return bool(
        section.section_mem_map.value_()
        & _section_flags(section.prog_)["SECTION_HAS_MEM_MAP"]
    )


@takes_program_or_default
def valid_section_nr(prog: Program, nr: IntegerLike) -> bool:
    """
    Return whether the SPARSEMEM section with the given number is valid, i.e.,
    has a ``mem_map``.

    :param nr: ``unsigned long``
    """
    return valid_section(nr_to_section(prog, nr))


def online_section(section: Object) -> bool:
    """
    Return whether a SPARSEMEM section is online.

    This is only valid since Linux kernel 4.13.

    :param section: ``struct mem_section *``
    """
    section = section.read_()
    if not section:
        return False
    return bool(
        section.section_mem_map.value_()
        & _section_flags(section.prog_)["SECTION_IS_ONLINE"]
    )


@takes_program_or_default
def online_section_nr(prog: Program, nr: IntegerLike) -> bool:
    """
    Return whether the SPARSEMEM section with the given number is online.

    This is only valid since Linux kernel 4.13.

    :param nr: ``unsigned long``
    """
    return online_section(nr_to_section(prog, nr))


def early_section(section: Object) -> bool:
    """
    Return whether a SPARSEMEM section was created during early memory
    initialization.

    This is only valid since Linux kernel 5.3.

    :param section: ``struct mem_section *``
    """
    section = section.read_()
    if not section:
        return False
    return bool(
        section.section_mem_map.value_()
        & _section_flags(section.prog_)["SECTION_IS_EARLY"]
    )


@takes_program_or_default
def early_section_nr(prog: Program, nr: IntegerLike) -> bool:
    """
    Return whether the SPARSEMEM section with the given number was created
    during early memory initialization.

    This is only valid since Linux kernel 5.3.

    :param nr: ``unsigned long``
    """
    return early_section(nr_to_section(prog, nr))


def decode_section_flags(section: Object) -> str:
    """
    Get a human-readable representation of the flags set on a SPARSEMEM section.

    >>> decode_section_flags(section)
    'SECTION_MARKED_PRESENT|SECTION_HAS_MEM_MAP|SECTION_IS_ONLINE|SECTION_IS_EARLY'

    :param section: ``struct mem_section *``
    """
    flags = _section_flags(section.prog_)
    return decode_flags(
        section.section_mem_map.value_() & ~flags["SECTION_MAP_MASK"],
        flags.items(),
        False,
    )


def _highest_present_section_nr(prog: Program) -> int:
    try:
        return prog["__highest_present_section_nr"].value_()
    except ObjectNotFoundError:
        pass
    # Before Linux kernel commit c4e1be9ec113 ("mm, sparsemem: break out of
    # loops early") (in v4.13), __highest_present_section_nr didn't exist, so
    # we have to loop all the way up to NR_MEM_SECTIONS.
    # NR_MEM_SECTIONS = 1 << SECTIONS_SHIFT
    # SECTIONS_SHIFT = MAX_PHYSMEM_BITS - SECTION_SIZE_BITS
    return 1 << (prog["MAX_PHYSMEM_BITS"] - prog["SECTION_SIZE_BITS"]).value_()


@takes_program_or_default
def for_each_present_section(prog: Program) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over each present SPARSEMEM section.

    :return: Iterator of (section number, ``struct mem_section *`` object)
        tuples.
    """
    SECTIONS_PER_ROOT = prog["SECTIONS_PER_ROOT"].value_()
    SECTION_MARKED_PRESENT = _section_flags(prog)["SECTION_MARKED_PRESENT"]

    highest_present_section_nr = _highest_present_section_nr(prog)
    nr_roots = highest_present_section_nr // SECTIONS_PER_ROOT + 1

    mem_section = prog["mem_section"]
    unaliased_type = mem_section.type_.unaliased()
    if unaliased_type.kind == TypeKind.POINTER:
        mem_section = mem_section.read_()
        if not mem_section:
            return

    root_kind = unaliased_type.type.unaliased_kind()

    nr = 0
    for root_nr, root in enumerate(mem_section[:nr_roots]):
        if root_kind == TypeKind.POINTER:
            root = root.read_()
            if not root:
                nr += SECTIONS_PER_ROOT
                continue

        if root_nr == nr_roots - 1:
            nr_sections = highest_present_section_nr % SECTIONS_PER_ROOT + 1
        else:
            nr_sections = SECTIONS_PER_ROOT
        for section in root[:nr_sections]:
            # Open-coded present_section() to avoid some overhead.
            if section.section_mem_map.value_() & SECTION_MARKED_PRESENT:
                yield nr, section.address_of_()
            nr += 1



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash timer command."""

import argparse
from typing import Any, List, Sequence

from drgn import FaultError, Object, Program, ProgramFlags
from drgn.commands import argument, drgn_argument
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    crash_command,
    parse_cpuspec,
)
from drgn.helpers.common.format import CellFormat, RowOptions, print_table
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.timekeeping import ktime_get_coarse_ns, ktime_to_ns
from drgn.helpers.linux.timer import (
    hrtimer_clock_base_for_each,
    timer_base_for_each,
    timer_base_names,
)


def _function_cell(function: Object) -> str:
    address = function.value_()
    try:
        function_symbol = function.prog_.symbol(address)
    except LookupError:
        return f"{address:x}"
    else:
        return f"{address:x}  <{function_symbol.name}>"


@crash_command(
    description="list kernel timers",
    arguments=(
        argument(
            "-r",
            dest="hrtimer",
            action="store_true",
            help="display high-resolution timers (hrtimers)",
        ),
        argument(
            "-C",
            dest="cpu",
            default="all",
            help="restrict the output to one or more CPUs, "
            "which may be a comma-separated string of CPU numbers or ranges "
            "(e.g., '0,3-4')",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_timer(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    **kwargs: Any,
) -> None:
    cpuspec = parse_cpuspec(args.cpu)

    if args.hrtimer:
        if args.drgn:
            code = CrashDrgnCodeBuilder(prog)
            code.add_from_import("drgn.helpers.linux.percpu", "per_cpu")
            code.add_from_import(
                "drgn.helpers.linux.timer", "hrtimer_clock_base_for_each"
            )
            code.add_from_import(
                "drgn.helpers.linux.timekeeping", "ktime_get_coarse_ns", "ktime_to_ns"
            )
            code.append("now = ktime_get_coarse_ns()\n\n")
            with code.begin_cpuspec_loop(cpuspec), code.begin_block(
                """\
cpu_base = per_cpu(prog["hrtimer_bases"], cpu)

for clock_base in cpu_base.clock_base:
"""
            ):
                code.append(
                    """\
clock = clock_base.index
current = now + ktime_to_ns(clock_base.offset)

"""
                )
                with code.begin_retry_loop_if_live(1000):
                    code.append(
                        """\
for hrtimer in hrtimer_clock_base_for_each(clock_base.address_of_()):
    softexpires = ktime_to_ns(hrtimer._softexpires)
    expires = ktime_to_ns(hrtimer.node.expires)
    tte = expires - current
    function = hrtimer.function
"""
                    )
            return code.print()

        hrtimer_bases = prog["hrtimer_bases"]
        now = ktime_get_coarse_ns(prog)
        first_cpu = True
        for cpu in cpuspec.cpus(prog):
            if first_cpu:
                first_cpu = False
            else:
                print()
            cpu_base = per_cpu(hrtimer_bases, cpu)
            print(f"CPU: {cpu}  HRTIMER_CPU_BASE: {cpu_base.address_:x}")

            first_clock_base = True
            for clock_base in cpu_base.clock_base:
                if first_clock_base:
                    first_clock_base = False
                else:
                    print()

                index = clock_base.index.read_()
                # Crash shows the name of the hrtimer_clock_base::get_time
                # function, but that was removed in Linux kernel commit
                # 009eb5da29a9 ("hrtimer: Remove hrtimer_clock_base::
                # Get_time") (in v6.18), so we omit it.
                print(
                    f"  CLOCK: {index.value_()}  HRTIMER_CLOCK_BASE: {clock_base.address_:x}"
                )

                current = (now + ktime_to_ns(clock_base.offset)).value_()
                rows: List[Sequence[Any]] = [
                    (
                        "",
                        CellFormat("CURRENT", "^"),
                    ),
                    ("", current),
                    (
                        "",
                        CellFormat("SOFTEXPIRES", "^"),
                        CellFormat("EXPIRES", "^"),
                        CellFormat("TTE", "^"),
                        CellFormat("HRTIMER", "^"),
                        "FUNCTION",
                    ),
                ]
                # Walking the hrtimer queue is racy. Retry a limited number of
                # times on live kernels.
                for _ in range(1000 if (prog.flags & ProgramFlags.IS_LIVE) else 1):
                    try:
                        for hrtimer in hrtimer_clock_base_for_each(clock_base):
                            expires = ktime_to_ns(hrtimer.node.expires).value_()
                            rows.append(
                                (
                                    "",
                                    ktime_to_ns(hrtimer._softexpires).value_(),
                                    expires,
                                    expires - current,
                                    CellFormat(hrtimer.value_(), "^x"),
                                    _function_cell(hrtimer.function),
                                ),
                            )
                        break
                    except FaultError:
                        del rows[3:]
                else:
                    print("  (corrupted)")
                    continue
                if len(rows) > 3:
                    # We print each base separately instead of as one big table
                    # because the timestamp values can differ greatly between
                    # bases and make the formatting look funny if they're all
                    # aligned to the largest one.
                    print_table(rows)
                else:
                    print("  (empty)")
    else:
        if args.drgn:
            code = CrashDrgnCodeBuilder(prog)
            code.append('jiffies = prog["jiffies"]\n\n')
            code.add_from_import("drgn.helpers.linux.percpu", "per_cpu")
            code.add_from_import(
                "drgn.helpers.linux.timer", "timer_base_for_each", "timer_base_names"
            )
            with code.begin_cpuspec_loop(cpuspec), code.begin_block(
                'for name, base in zip(timer_base_names(), per_cpu(prog["timer_bases"], cpu)):\n'
            ), code.begin_retry_loop_if_live(1000):
                code.append(
                    """\
for timer in timer_base_for_each(base):
    expires = timer.expires
    tte = expires - jiffies
    function = timer.function
"""
                )
            return code.print()

        jiffies = prog["jiffies"].value_()
        print(f"JIFFIES\n{jiffies}")

        rows = []
        timer_bases = prog["timer_bases"]
        for cpu in cpuspec.cpus(prog):
            rows.append(())
            for name, base in zip(timer_base_names(prog), per_cpu(timer_bases, cpu)):
                rows.append(
                    RowOptions(
                        (f"TIMER_BASES[{cpu}][{name}]: {base.address_:x}",),
                        group=1,
                    ),
                )

                rows.append(
                    (
                        "",
                        "EXPIRES",
                        CellFormat("TTE", ">"),
                        CellFormat("TIMER_LIST", "^"),
                        "FUNCTION",
                    ),
                )
                timer_rows = []
                # Walking the timer lists is racy. Retry a limited number of
                # times on live kernels.
                for _ in range(1000 if (prog.flags & ProgramFlags.IS_LIVE) else 1):
                    try:
                        for timer in timer_base_for_each(base):
                            expires = timer.expires.value_()
                            timer_rows.append(
                                (
                                    "",
                                    CellFormat(expires, "<"),
                                    expires - jiffies,
                                    CellFormat(timer.value_(), "^x"),
                                    _function_cell(timer.function),
                                )
                            )
                        break
                    except FaultError:
                        timer_rows.clear()
                else:
                    rows.append(("", "(corrupted)"))
                    continue
                if timer_rows:
                    rows.extend(timer_rows)
                else:
                    rows.append(("", "(none)"))
        print_table(rows)
